from __future__ import annotations

import operator
from operator import methodcaller

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from pytest import param

import xorq.api as xo
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.api import _
from xorq.backends.pandas import Backend
from xorq.tests.util import assert_frame_equal, assert_series_equal
from xorq.vendor import ibis


def test_table_column(t, df):
    expr = t.plain_int64
    result = expr.execute()
    expected = df.plain_int64
    assert_series_equal(result, expected)


def test_literal(client):
    assert client.execute(ibis.literal(1)) == 1


def test_selection(t, df):
    expr = t[((t.plain_strings == "a") | (t.plain_int64 == 3)) & (t.dup_strings == "d")]
    result = expr.execute()
    expected = df[
        ((df.plain_strings == "a") | (df.plain_int64 == 3)) & (df.dup_strings == "d")
    ].reset_index(drop=True)

    assert_frame_equal(result[expected.columns], expected)


def test_mutate(t, df):
    expr = t.mutate(x=t.plain_int64 + 1, y=t.plain_int64 * 2)
    result = expr.execute()
    expected = df.assign(x=df.plain_int64 + 1, y=df.plain_int64 * 2)
    assert_frame_equal(result[expected.columns], expected)


def test_project_scope_does_not_override(t, df):
    col = t.plain_int64
    expr = t[
        [
            col.name("new_col"),
            col.sum().over(ibis.window(group_by="dup_strings")).name("grouped"),
        ]
    ]
    result = expr.execute()
    expected = pd.concat(
        [
            df[["plain_int64", "dup_strings"]].rename(
                columns={"plain_int64": "new_col"}
            ),
            df.groupby("dup_strings")
            .plain_int64.transform("sum")
            .reset_index(drop=True)
            .rename("grouped"),
        ],
        axis=1,
    )[["new_col", "grouped"]]
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "where",
    [
        param(lambda _: None, id="none"),
        param(lambda t: t.dup_strings == "d", id="simple"),
        param(lambda t: (t.dup_strings == "d") | (t.plain_int64 < 100), id="complex"),
    ],
)
@pytest.mark.parametrize(
    ("ibis_func", "pandas_func"),
    [
        param(methodcaller("abs"), np.abs, id="abs"),
        param(methodcaller("ceil"), np.ceil, id="ceil"),
        param(methodcaller("exp"), np.exp, id="exp"),
        param(methodcaller("floor"), np.floor, id="floor"),
        param(methodcaller("ln"), np.log, id="log"),
        param(methodcaller("log10"), np.log10, id="log10"),
        param(methodcaller("log", 2), lambda x: np.log(x) / np.log(2), id="logb"),
        param(methodcaller("log2"), np.log2, id="log2"),
        param(
            methodcaller("round", 0), lambda x: x.round(0).astype("int64"), id="round0"
        ),
        param(methodcaller("round", -2), methodcaller("round", -2), id="roundm2"),
        param(methodcaller("round", 2), methodcaller("round", 2), id="round2"),
        param(methodcaller("round"), lambda x: x.round().astype("int64"), id="round"),
        param(methodcaller("sign"), np.sign, id="sign"),
        param(methodcaller("sqrt"), np.sqrt, id="sqrt"),
    ],
)
def test_aggregation_group_by(t, df, where, ibis_func, pandas_func):
    ibis_where = where(t)
    expr = t.group_by(t.dup_strings).aggregate(
        avg_plain_int64=t.plain_int64.mean(where=ibis_where),
        sum_plain_float64=t.plain_float64.sum(where=ibis_where),
        mean_float64_positive=ibis_func(t.float64_positive).mean(where=ibis_where),
        neg_mean_int64_with_zeros=(-t.int64_with_zeros).mean(where=ibis_where),
        nunique_dup_ints=t.dup_ints.nunique(),
    )
    result = expr.execute()

    pandas_where = where(df)
    mask = slice(None) if pandas_where is None else pandas_where
    expected = (
        df.groupby("dup_strings")
        .agg(
            {
                "plain_int64": lambda x, mask=mask: x[mask].mean(),
                "plain_float64": lambda x, mask=mask: x[mask].sum(),
                "dup_ints": "nunique",
                "float64_positive": (
                    lambda x, mask=mask, func=pandas_func: func(x[mask]).mean()
                ),
                "int64_with_zeros": lambda x, mask=mask: (-x[mask]).mean(),
            }
        )
        .reset_index()
        .rename(
            columns={
                "plain_int64": "avg_plain_int64",
                "plain_float64": "sum_plain_float64",
                "dup_ints": "nunique_dup_ints",
                "float64_positive": "mean_float64_positive",
                "int64_with_zeros": "neg_mean_int64_with_zeros",
            }
        )
    )
    lhs = result[expected.columns]
    rhs = expected
    assert_frame_equal(lhs, rhs)


def test_aggregation_without_group_by(t, df):
    expr = t.aggregate(
        avg_plain_int64=t.plain_int64.mean(),
        sum_plain_float64=t.plain_float64.sum(),
    )
    result = expr.execute()[["avg_plain_int64", "sum_plain_float64"]]
    new_names = {
        "plain_float64": "sum_plain_float64",
        "plain_int64": "avg_plain_int64",
    }
    expected = (
        pd.Series(
            [df["plain_int64"].mean(), df["plain_float64"].sum()],
            index=["plain_int64", "plain_float64"],
        )
        .to_frame()
        .T.rename(columns=new_names)
    )
    assert_frame_equal(result[expected.columns], expected)


def test_group_by_with_having(t, df):
    expr = (
        t.group_by(t.dup_strings)
        .having(t.plain_float64.sum() == 5)
        .aggregate(avg_a=t.plain_int64.mean(), sum_c=t.plain_float64.sum())
    )
    result = expr.execute()

    expected = (
        df.groupby("dup_strings")
        .agg({"plain_int64": "mean", "plain_float64": "sum"})
        .reset_index()
        .rename(columns={"plain_int64": "avg_a", "plain_float64": "sum_c"})
    )
    expected = expected.loc[expected.sum_c == 5, ["avg_a", "sum_c"]]

    assert_frame_equal(result[expected.columns], expected)


def test_group_by_rename_key(t, df):
    expr = t.group_by(t.dup_strings.name("foo")).aggregate(
        dup_string_count=t.dup_strings.count()
    )
    assert "foo" in expr.schema()
    result = expr.execute()
    assert "foo" in result.columns

    expected = (
        df.groupby("dup_strings")
        .dup_strings.count()
        .rename("dup_string_count")
        .reset_index()
        .rename(columns={"dup_strings": "foo"})
    )
    assert_frame_equal(result, expected)


@pytest.mark.parametrize("reduction", ["mean", "sum", "count", "std", "var"])
@pytest.mark.parametrize(
    "where",
    [
        lambda t: (t.plain_strings == "a") | (t.plain_strings == "c"),
        lambda t: (
            (t.dup_strings == "d") & ((t.plain_int64 == 1) | (t.plain_int64 == 3))
        ),
        lambda t: None,
    ],
)
def test_reduction(t, df, reduction, where):
    func = getattr(t.plain_int64, reduction)
    mask = where(t)
    expr = func(where=mask)
    result = expr.execute()

    df_mask = where(df)
    expected_func = getattr(
        df.loc[df_mask if df_mask is not None else slice(None), "plain_int64"],
        reduction,
    )
    expected = expected_func()
    assert result == expected


@pytest.mark.parametrize(
    "reduction",
    [
        lambda x: x.any(),
        lambda x: x.all(),
        lambda x: ~(x.any()),
        lambda x: ~(x.all()),
    ],
)
def test_boolean_aggregation(t, df, reduction):
    expr = reduction(t.plain_int64 == 1)
    result = expr.execute()
    expected = reduction(df.plain_int64 == 1)
    assert result == expected


@pytest.mark.parametrize("column", ["float64_with_zeros", "int64_with_zeros"])
def test_nullif_zero(t, df, column):
    expr = t[column].nullif(0)
    result = expr.execute()
    expected = df[column].replace(0, np.nan)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("left", "right", "expected", "compare"),
    [
        param(
            lambda t: ibis.literal(1),
            lambda t: ibis.literal(1),
            lambda df: np.nan,
            np.testing.assert_array_equal,  # treats NaNs as equal
            id="literal_literal_equal",
        ),
        param(
            lambda t: ibis.literal(1),
            lambda t: ibis.literal(2),
            lambda df: 1,
            np.testing.assert_equal,
            id="literal_literal_not_equal",
        ),
        param(
            lambda t: t.dup_strings,
            lambda t: ibis.literal("a"),
            lambda df: df.dup_strings.where(df.dup_strings != "a"),
            assert_series_equal,
            id="series_literal",
        ),
        param(
            lambda t: t.dup_strings,
            lambda t: t.dup_strings,
            lambda df: df.dup_strings.where(df.dup_strings != df.dup_strings),
            assert_series_equal,
            id="series_series",
        ),
        param(
            lambda t: ibis.literal("a"),
            lambda t: t.dup_strings,
            lambda _: pd.Series(["a", np.nan, "a"], name="dup_strings"),
            assert_series_equal,
            id="literal_series",
        ),
    ],
)
def test_nullif(t, df, left, right, expected, compare):
    expr = left(t).nullif(right(t))
    result = Backend().execute(expr)
    compare(result, expected(df))


def test_nullif_inf():
    df = pd.DataFrame({"a": [np.inf, 3.14, -np.inf, 42.0]})
    con = Backend().connect({"t": df})
    t = con.table("t")
    expr = t.a.nullif(np.inf).nullif(-np.inf)
    result = expr.execute()
    expected = pd.Series([np.nan, 3.14, np.nan, 42.0], name="a")
    assert_series_equal(result, expected)


def test_group_concat(t, df):
    expr = t.group_by(t.dup_strings).aggregate(foo=t.plain_int64.group_concat(","))
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .apply(lambda df: ",".join(df.plain_int64.astype(str)))
        .reset_index()
        .rename(columns={0: "foo"})
    )
    assert_frame_equal(result[expected.columns], expected)


@pytest.mark.parametrize("offset", [0, 2])
def test_frame_limit(t, df, offset):
    n = 5
    df_expr = t.limit(n, offset=offset)
    result = df_expr.execute()
    expected = df.iloc[offset : offset + n].reset_index(drop=True)
    assert_frame_equal(result[expected.columns], expected)


@pytest.mark.parametrize("offset", [0, 2])
def test_series_limit(t, df, offset):
    with pytest.raises(AttributeError):
        t.plain_int64.limit(5, offset=offset)


@pytest.mark.parametrize(
    ("key", "pandas_by", "pandas_ascending"),
    [
        (lambda t, col: [ibis.desc(t[col])], lambda col: [col], False),
        (
            lambda t, col: [t[col], ibis.desc(t.plain_int64)],
            lambda col: [col, "plain_int64"],
            [True, False],
        ),
        (
            lambda t, col: [ibis.desc(t.plain_int64 * 2)],
            lambda col: ["plain_int64"],
            False,
        ),
    ],
)
@pytest.mark.parametrize(
    "column",
    ["plain_datetimes_naive", "plain_datetimes_ny", "plain_datetimes_utc"],
)
def test_order_by(t, df, column, key, pandas_by, pandas_ascending):
    expr = t.order_by(key(t, column))
    result = expr.execute()
    expected = df.sort_values(
        pandas_by(column), ascending=pandas_ascending
    ).reset_index(drop=True)
    assert_frame_equal(result[expected.columns], expected)


def test_complex_order_by(t, df):
    expr = t.order_by([ibis.desc(t.plain_int64 * t.plain_float64), t.plain_float64])
    result = expr.execute()
    expected = (
        df.assign(foo=df.plain_int64 * df.plain_float64)
        .sort_values(["foo", "plain_float64"], ascending=[False, True])
        .drop(["foo"], axis=1)
        .reset_index(drop=True)
    )

    assert_frame_equal(result[expected.columns], expected)


def test_count_distinct(t, df):
    expr = t.dup_strings.nunique()
    result = expr.execute()
    expected = df.dup_strings.nunique()
    assert result == expected


def test_value_counts(t, df):
    expr = t.dup_strings.value_counts()
    result = expr.execute()
    expected = (
        df.dup_strings.value_counts()
        .rename("dup_strings")
        .reset_index(name="dup_strings_count")
        .rename(columns={"index": "dup_strings"})
        .sort_values(["dup_strings"])
        .reset_index(drop=True)
    )
    assert_frame_equal(result[expected.columns], expected)


def test_table_count(t, df):
    expr = t.count()
    result = expr.execute()
    expected = len(df)
    assert result == expected


def test_weighted_average(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        avg=(t.plain_float64 * t.plain_int64).sum() / t.plain_int64.sum()
    )
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .apply(
            lambda df: (df.plain_int64 * df.plain_float64).sum() / df.plain_int64.sum()
        )
        .reset_index()
        .rename(columns={0: "avg"})
    )
    assert_frame_equal(result[expected.columns], expected)


def test_group_by_multiple_keys(t, df):
    expr = t.group_by([t.dup_strings, t.dup_ints]).aggregate(
        avg_plain_float64=t.plain_float64.mean()
    )
    result = expr.execute()
    expected = (
        df.groupby(["dup_strings", "dup_ints"])
        .agg({"plain_float64": "mean"})
        .reset_index()
        .rename(columns={"plain_float64": "avg_plain_float64"})
    )
    assert_frame_equal(result[expected.columns], expected)


def test_mutate_after_group_by(t, df):
    gb = t.group_by(t.dup_strings).aggregate(avg_plain_float64=t.plain_float64.mean())
    expr = gb.mutate(x=gb.avg_plain_float64)
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .agg({"plain_float64": "mean"})
        .reset_index()
        .rename(columns={"plain_float64": "avg_plain_float64"})
    )
    expected = expected.assign(x=expected.avg_plain_float64)
    assert_frame_equal(result[expected.columns], expected)


def test_groupby_with_unnamed_arithmetic(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        naive_variance=((t.plain_float64**2).sum() - t.plain_float64.mean() ** 2)
        / t.plain_float64.count()
    )
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .agg({"plain_float64": lambda x: ((x**2).sum() - x.mean() ** 2) / x.count()})
        .reset_index()
        .rename(columns={"plain_float64": "naive_variance"})
    )
    assert_frame_equal(result[expected.columns], expected)


def test_isnull(t, df):
    expr = t.strings_with_nulls.isnull()
    result = expr.execute()
    expected = df.strings_with_nulls.isnull()
    assert_series_equal(result, expected)


def test_notnull(t, df):
    expr = t.strings_with_nulls.notnull()
    result = expr.execute()
    expected = df.strings_with_nulls.notnull()
    assert_series_equal(result, expected)


@pytest.mark.parametrize("raw_value", [0.0, 1.0])
def test_scalar_parameter(t, df, raw_value):
    value = ibis.param(dt.double)
    expr = t.float64_with_zeros == value
    result = expr.execute(params={value: raw_value})
    expected = df.float64_with_zeros == raw_value
    assert_series_equal(result, expected)


@pytest.mark.parametrize("elements", [[1], (1,), {1}, frozenset({1})])
def test_isin(t, df, elements):
    expr = t.plain_float64.isin(elements)
    expected = df.plain_float64.isin(elements)
    result = expr.execute()
    assert_series_equal(result, expected)


@pytest.mark.parametrize("elements", [[1], (1,), {1}, frozenset({1})])
def test_notin(t, df, elements):
    expr = t.plain_float64.notin(elements)
    expected = ~df.plain_float64.isin(elements)
    result = expr.execute()
    assert_series_equal(result, expected)


def test_cast_on_group_by(t, df):
    expr = t.group_by(t.dup_strings).aggregate(
        casted=(t.float64_with_zeros == 0).cast("int64").sum()
    )
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .float64_with_zeros.apply(lambda s: (s == 0).astype("int64").sum())
        .reset_index()
        .rename(columns={"float64_with_zeros": "casted"})
    )
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "opname", ["add", "mul", "sub", "truediv", "floordiv", "mod", "pow"]
)
@pytest.mark.parametrize(
    "argfunc", [param(lambda c: (1.0, c), id="1c"), param(lambda c: (c, 1.0), id="c1")]
)
def test_left_binary_op(t, df, opname, argfunc):
    op = getattr(operator, opname)
    left, right = argfunc(t.float64_with_zeros)
    expr = op(left, right)
    result = expr.execute()
    expected = op(*argfunc(df.float64_with_zeros))
    assert_series_equal(
        result,
        expected,
        check_dtype=not (isinstance(right, float) and opname == "floordiv"),
    )


@pytest.mark.parametrize(
    "opname", ["add", "mul", "sub", "truediv", "floordiv", "mod", "pow"]
)
@pytest.mark.parametrize(
    "argfunc", [param(lambda c: (1.0, c), id="1c"), param(lambda c: (c, 1.0), id="c1")]
)
def test_left_binary_op_gb(t, df, opname, argfunc):
    op = getattr(operator, opname)
    left, right = argfunc(t.float64_with_zeros)
    expr = t.group_by("dup_strings").aggregate(foo=op(left, right).sum())
    result = expr.execute()
    expected = (
        df.groupby("dup_strings")
        .float64_with_zeros.apply(lambda s: op(*argfunc(s)).sum())
        .reset_index()
        .rename(columns={"float64_with_zeros": "foo"})
    )
    assert_frame_equal(
        result,
        expected,
        check_dtype=not (isinstance(right, float) and opname == "floordiv"),
    )


@pytest.mark.parametrize(
    "left_f",
    [
        param(lambda e: e - 1, id="sub"),
        param(lambda _: 0.0, id="zero"),
        param(lambda _: None, id="none"),
    ],
)
@pytest.mark.parametrize(
    "right_f",
    [
        param(lambda e: e + 1, id="add"),
        param(lambda _: 1.0, id="one"),
        param(lambda _: None, id="none"),
    ],
)
def test_ifelse_series(t, df, left_f, right_f):
    col_expr = t["plain_int64"]
    result = ibis.ifelse(
        col_expr > col_expr.mean(), left_f(col_expr), right_f(col_expr)
    ).execute()

    series = df["plain_int64"]
    cond = series > series.mean()
    left = left_f(series)
    if not isinstance(left, pd.Series):
        left = pd.Series(np.repeat(left, len(cond)), name=cond.name)
    expected = left.where(cond, right_f(series))

    assert_series_equal(
        result.astype(object).fillna(pd.NA),
        expected.astype(object).fillna(pd.NA),
        check_dtype=False,
    )


@pytest.mark.parametrize(
    ("cond", "expected_func"),
    [
        param(True, lambda df: df["plain_int64"].astype("float64"), id="true"),
        param(False, lambda df: pd.Series(np.repeat(3.0, len(df))), id="false"),
    ],
)
def test_ifelse_scalar(t, df, cond, expected_func):
    expr = ibis.ifelse(cond, t["plain_int64"], 3.0)
    result = expr.execute()
    expected = expected_func(df)
    assert_series_equal(result, expected)


def test_ifelse_long(batting, batting_df):
    col_expr = batting["AB"]
    result = ibis.ifelse(col_expr > col_expr.mean(), col_expr, 0.0).execute()

    series = batting_df["AB"]
    expected = series.where(series > series.mean(), other=0.0).astype("float64")

    assert_series_equal(result, expected)


def test_round(t, df):
    precision = 2
    mult = 3.33333
    result = (t.count() * mult).round(precision).execute()
    expected = np.around(len(df) * mult, precision)
    npt.assert_almost_equal(result, expected, decimal=precision)


def test_quantile_groupby(batting, batting_df):
    def q_fun(x, quantile):
        res = x.quantile(quantile).tolist()
        return [res for _ in range(len(x))]

    frac = 0.2
    result = (
        batting.group_by("teamID")
        .mutate(res=lambda x: x.RBI.quantile([frac, 1 - frac]))
        .res.execute()
    )
    expected = (
        batting_df.groupby("teamID")
        .RBI.transform(q_fun, quantile=[frac, 1 - frac])
        .rename("res")
    )
    assert_series_equal(result, expected)


def test_summary_numeric(batting, batting_df):
    expr = batting.aggregate(
        count=_.G.count(),
        nulls=_.G.isnull().sum(),
        min=_.G.min(),
        max=_.G.max(),
        sum=_.G.sum(),
        mean=_.G.mean(),
        approx_nunique=_.G.nunique(),
    )
    result = expr.execute()
    assert len(result) == 1

    G = batting_df.G
    expected = {
        "count": G.count(),
        "nulls": G.isnull().sum(),
        "min": G.min(),
        "max": G.max(),
        "sum": G.sum(),
        "mean": G.mean(),
        "approx_nunique": G.nunique(),
    }
    assert dict(result.iloc[0]) == expected


def test_summary_non_numeric(batting, batting_df):
    expr = batting.aggregate(
        count=_.teamID.count(),
        nulls=_.teamID.isnull().sum(),
        uniques=_.teamID.nunique(),
    )
    result = expr.execute()
    assert len(result) == 1
    assert len(result.columns) == 3
    expected = {
        "count": batting_df.teamID.count(),
        "nulls": batting_df.teamID.isnull().sum(),
        "uniques": batting_df.teamID.nunique(),
    }
    assert dict(result.iloc[0]) == expected


def test_non_range_index():
    def do_replace(col):
        return col.cases(
            *(
                (1, "one"),
                (2, "two"),
            ),
            else_="unk",
        )

    df = pd.DataFrame(
        {
            "A": pd.Series({i: i % 3 for i in (0, 1, 2, 4)}),
            "B": 0,
        }
    )
    expr = (
        xo.pandas.connect({"t": df})
        .table("t")
        .mutate(A=lambda t: t["A"].pipe(do_replace))
    )
    assert df.index.equals(expr.execute().index)


def test_table_distinct(t, df):
    expr = t[["dup_strings"]].distinct()
    result = expr.execute()
    expected = df[["dup_strings"]].drop_duplicates()
    assert_frame_equal(result, expected)


@pytest.mark.parametrize("distinct", [True, False])
def test_union(client, df1, distinct):
    t = client.table("df1")
    expr = t.union(t, distinct=distinct)
    result = expr.execute()
    expected = df1 if distinct else pd.concat([df1, df1], axis=0, ignore_index=True)
    assert_frame_equal(result, expected)


def test_intersect(client, df1, intersect_df2):
    t1 = client.table("df1")
    t2 = client.table("intersect_df2")
    expr = t1.intersect(t2)
    result = expr.execute()
    expected = df1.merge(intersect_df2, on=list(df1.columns))
    assert_frame_equal(result, expected)


def test_difference(client, df1, intersect_df2):
    t1 = client.table("df1")
    t2 = client.table("intersect_df2")
    expr = t1.difference(t2)
    result = expr.execute()
    merged = df1.merge(intersect_df2, on=list(df1.columns), how="outer", indicator=True)
    expected = merged[merged["_merge"] != "both"].drop("_merge", axis=1)
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "distinct",
    [
        param(
            True,
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason="Pandas cannot compute the distinct element of an array column",
            ),
        ),
        False,
    ],
)
def test_union_with_list_types(t, df, distinct):
    expr = t.union(t, distinct=distinct)
    result = expr.execute()
    expected = df if distinct else pd.concat([df, df], axis=0, ignore_index=True)
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "operation",
    [
        pytest.param(
            lambda column: column + column,
            id="sum",
        ),
        pytest.param(
            lambda column: column + 1,
            id="sum_scalar",
        ),
        pytest.param(
            lambda column: column - column,
            id="subtract",
        ),
        pytest.param(
            lambda column: column - 1,
            id="subtract_scalar",
        ),
        pytest.param(
            lambda column: column * column,
            id="multiply",
        ),
        pytest.param(
            lambda column: column * 2,
            id="multiply_scalar",
        ),
        pytest.param(
            lambda column: column % column,
            id="mod",
            marks=pytest.mark.xfail(
                raises=ZeroDivisionError,
                reason=("Ibis cannot modulo divide two unsigned integer columns."),
            ),
        ),
        pytest.param(
            lambda column: column % 2,
            id="mod_scalar",
        ),
        pytest.param(
            lambda column: column / column,
            id="divide",
        ),
        pytest.param(
            lambda column: column / 2,
            id="divide_scalar",
        ),
        pytest.param(
            lambda column: column // column,
            id="floordivide",
        ),
        pytest.param(
            lambda column: column // 2,
            id="floordivide_scalar",
        ),
    ],
)
def test_unsigned_integers(t, df, operation):
    expr = operation(t.plain_uint64)
    result = expr.execute()
    expected = operation(df.plain_uint64)
    expected_dtype = expr.type().to_pandas()
    assert result.dtype == expected_dtype
    assert_series_equal(result, expected.astype(expected_dtype))



import xorq.vendor.ibis as ibis
from xorq.ibis_yaml.common import RefEnum, RegistryEnum


def test_window_function_roundtrip(compiler, t):
    expr = t.select(
        [
            t.c.mean()
            .over(ibis.window(preceding=5, following=0, group_by=t.a))
            .name("mean_c")
        ]
    )

    yaml_dict = compiler.to_yaml(expr)

    reconstructed_expr = compiler.from_yaml(yaml_dict)

    assert expr.equals(reconstructed_expr)


def test_aggregation_window(compiler, t):
    cases = [
        (None, None),
        (5, 0),
        (0, 5),
        (5, 5),
    ]

    for preceding, following in cases:
        expr = t.select(
            [
                t.c.mean()
                .over(
                    ibis.window(preceding=preceding, following=following, group_by=t.a)
                )
                .name("mean_c")
            ]
        )

        yaml_dict = compiler.to_yaml(expr)
        node_ref = yaml_dict["expression"][RefEnum.node_ref]

        expression = yaml_dict["definitions"][RegistryEnum.nodes][node_ref]
        assert expression["op"] == "Project"
        window_func = expression["values"]["mean_c"]
        assert window_func["op"] == "WindowFunction"
        assert window_func["args"][0]["op"] == "Mean"

        if preceding is None:
            assert "start" not in window_func
        else:
            assert window_func["start"] == preceding

        if following is None:
            assert "end" not in window_func
        else:
            assert window_func["end"] == following


def test_row_number_simple_roundtrip(compiler, t):
    expr = t.select([ibis.row_number().name("row_num")])
    yaml_dict = compiler.to_yaml(expr)
    reconstructed_expr = compiler.from_yaml(yaml_dict)
    assert expr.equals(reconstructed_expr)


def test_row_number_window_roundtrip(compiler, t):
    expr = t.select(
        [
            ibis.row_number()
            .over(
                ibis.window(
                    group_by=[t.a, t.b],
                    order_by=[t.c.desc(), t.d],
                    preceding=5,
                    following=0,
                )
            )
            .name("row_num")
        ]
    )
    yaml_dict = compiler.to_yaml(expr)
    reconstructed_expr = compiler.from_yaml(yaml_dict)
    assert expr.equals(reconstructed_expr)


def test_multiple_rank_expressions_roundtrip(compiler, t):
    expr = t.select(
        [
            ibis.row_number().over(ibis.window(group_by=t.a)).name("simple_row_num"),
            ibis.row_number()
            .over(ibis.window(group_by=[t.a, t.b], order_by=t.c.desc()))
            .name("ordered_row_num"),
            t.c.mean()
            .over(ibis.window(preceding=3, following=0, group_by=t.a))
            .name("mean_c"),
        ]
    )
    yaml_dict = compiler.to_yaml(expr)
    reconstructed_expr = compiler.from_yaml(yaml_dict)
    assert expr.equals(reconstructed_expr)



"""PostgreSQL backend."""

from __future__ import annotations

import contextlib
import inspect
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import psycopg
import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.common.exceptions as exc
import xorq.vendor.ibis.backends.sql.compilers as sc
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends import (
    CanCreateDatabase,
    CanListCatalog,
)
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.backends.sql.compilers.base import TRUE, C, ColGen


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from typing_extensions import Self


class NatDumper(psycopg.adapt.Dumper):
    def dump(self, obj, context: Any | None = None) -> str | None:
        return None


class Backend(
    SQLBackend,
    CanListCatalog,
    CanCreateDatabase,
):
    name = "postgres"
    compiler = sc.postgres.compiler
    supports_python_udfs = True
    supports_temporary_tables = True

    def _from_url(self, url: ParseResult, **kwargs):
        """Connect to a backend using a URL `url`.

        Parameters
        ----------
        url
            URL with which to connect to a backend.
        kwargs
            Additional keyword arguments

        Returns
        -------
        BaseBackend
            A backend instance

        """
        database, *schema = url.path[1:].split("/", 1)
        connect_args = {
            "user": url.username,
            "password": unquote_plus(url.password or ""),
            "host": url.hostname,
            "database": database or "",
            "schema": schema[0] if schema else "",
            "port": url.port,
        }

        kwargs.update(connect_args)
        self._convert_kwargs(kwargs)

        if "user" in kwargs and not kwargs["user"]:
            del kwargs["user"]

        if "host" in kwargs and not kwargs["host"]:
            del kwargs["host"]

        if "database" in kwargs and not kwargs["database"]:
            del kwargs["database"]

        if "schema" in kwargs and not kwargs["schema"]:
            del kwargs["schema"]

        if "password" in kwargs and kwargs["password"] is None:
            del kwargs["password"]

        if "port" in kwargs and kwargs["port"] is None:
            del kwargs["port"]

        return self.connect(**kwargs)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise exc.XorqTypeError(
                f"{self.name} cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        name = op.name
        quoted = self.compiler.quoted
        create_stmt = sg.exp.Create(
            kind="TABLE",
            this=sg.exp.Schema(
                this=sg.to_identifier(name, quoted=quoted),
                expressions=schema.to_sqlglot(self.dialect),
            ),
            properties=sg.exp.Properties(expressions=[sge.TemporaryProperty()]),
        )
        create_stmt_sql = create_stmt.sql(self.dialect)

        table = op.data.to_pyarrow(schema)
        sql = self._build_insert_template(
            name, schema=schema, columns=True, placeholder="%s"
        )

        con = self.con
        with con.cursor() as cursor, con.transaction():
            cursor.execute(create_stmt_sql).executemany(
                sql, zip(*table.to_pydict().values())
            )

    @contextlib.contextmanager
    def begin(self):
        with (con := self.con).cursor() as cursor, con.transaction():
            yield cursor

    def execute(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | pd.Series | Any:
        """Execute an Ibis expression and return a pandas `DataFrame`, `Series`, or scalar.

        Parameters
        ----------
        expr
            Ibis expression to execute.
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            no limit. The default is in `ibis/config.py`.
        kwargs
            Keyword arguments

        Returns
        -------
        DataFrame | Series | scalar
            The result of the expression execution.
        """
        import pandas as pd
        from ibis.backends.postgres.converter import PostgresPandasData

        self._run_pre_execute_hooks(expr)

        table = expr.as_table()
        sql = self.compile(table, params=params, limit=limit, **kwargs)

        con = self.con
        with con.cursor() as cur, con.transaction():
            rows = cur.execute(sql).fetchall()

        schema = table.schema()
        df = pd.DataFrame.from_records(rows, columns=schema.names, coerce_float=True)
        df = PostgresPandasData.convert_table(df, schema)
        return expr.__pandas_result__(df)

    @property
    def version(self):
        version = f"{self.con.info.server_version:0>6}"
        major = int(version[:2])
        minor = int(version[2:4])
        patch = int(version[4:])
        pieces = [major]
        if minor:
            pieces.append(minor)
        pieces.append(patch)
        return ".".join(map(str, pieces))

    def do_connect(
        self,
        host: str | None = None,
        user: str | None = None,
        password: str | None = None,
        port: int = 5432,
        database: str | None = None,
        schema: str | None = None,
        autocommit: bool = True,
        **kwargs: Any,
    ) -> None:
        """Create an Ibis client connected to PostgreSQL database.

        Parameters
        ----------
        host
            Hostname
        user
            Username
        password
            Password
        port
            Port number
        database
            Database to connect to
        schema
            PostgreSQL schema to use. If `None`, use the default `search_path`.
        autocommit
            Whether or not to autocommit
        kwargs
            Additional keyword arguments to pass to the backend client connection.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_POSTGRES_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_POSTGRES_USER", "postgres")
        >>> password = os.environ.get("IBIS_TEST_POSTGRES_PASSWORD", "postgres")
        >>> database = os.environ.get("IBIS_TEST_POSTGRES_DATABASE", "ibis_testing")
        >>> con = ibis.postgres.connect(database=database, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int32
          bool_col        boolean
          tinyint_col     int16
          smallint_col    int16
          int_col         int32
          bigint_col      int64
          float_col       float32
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp(6)
          year            int32
          month           int32
        """
        import psycopg

        self.con = psycopg.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            dbname=database,
            autocommit=autocommit,
            **kwargs,
        )

        self._post_connect()

    @util.experimental
    @classmethod
    def from_connection(cls, con: psycopg.Connection, /) -> Backend:
        """Create an Ibis client from an existing connection to a PostgreSQL database.

        Parameters
        ----------
        con
            An existing connection to a PostgreSQL database.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect()
        return new_backend

    def _post_connect(self) -> None:
        import pandas as pd

        self.con.adapters.register_dumper(type(pd.NaT), NatDumper)
        with (con := self.con).cursor() as cursor, con.transaction():
            cursor.execute("SET TIMEZONE = UTC")
            if schema := self._con_kwargs.get("schema"):
                # The `false` means this setting will persist beyond this transaction
                cursor.execute("SELECT set_config('search_path', %s, false)", (schema,))

    @property
    def _session_temp_db(self) -> str | None:
        # Postgres doesn't assign the temporary table database until the first
        # temp table is created in a given session.
        # Before that temp table is created, this will return `None`
        # After a temp table is created, it will return `pg_temp_N` where N is
        # some integer
        con = self.con
        with con.cursor() as cursor, con.transaction():
            res = cursor.execute(
                "SELECT nspname FROM pg_namespace WHERE oid = pg_my_temp_schema()"
            ).fetchone()
        if res is not None:
            return res[0]
        return res

    def list_tables(
        self, *, like: str | None = None, database: tuple[str, str] | str | None = None
    ) -> list[str]:
        if database is not None:
            table_loc = database
        else:
            table_loc = (self.current_catalog, self.current_database)

        table_loc = self._to_sqlglot_table(table_loc)

        conditions = [TRUE]

        if (db := table_loc.args["db"]) is not None:
            db.args["quoted"] = False
            db = db.sql(dialect=self.name)
            conditions.append(C.table_schema.eq(sge.convert(db)))
        if (catalog := table_loc.args["catalog"]) is not None:
            catalog.args["quoted"] = False
            catalog = catalog.sql(dialect=self.name)
            conditions.append(C.table_catalog.eq(sge.convert(catalog)))

        sg_expr = (
            sg.select(C.table_name)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
            .where(*conditions)
        )

        # Include temporary tables only if no database has been explicitly
        # specified to avoid temp tables showing up in all calls to
        # `list_tables`
        if db == "public":
            # postgres temporary tables are stored in a separate schema so we need
            # to independently grab them and return them along with the existing
            # results
            sg_expr = sg_expr.union(
                sg.select(C.table_name)
                .from_(sg.table("tables", db="information_schema"))
                .distinct()
                .where(C.table_type.eq(sge.convert("LOCAL TEMPORARY"))),
                distinct=False,
            )

        sql = sg_expr.sql(self.dialect)
        con = self.con
        with con.cursor() as cursor, con.transaction():
            out = cursor.execute(sql).fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def list_catalogs(self, *, like: str | None = None) -> list[str]:
        # http://dba.stackexchange.com/a/1304/58517
        cats = (
            sg.select(C.datname)
            .from_(sg.table("pg_database", db="pg_catalog"))
            .where(sg.not_(C.datistemplate))
            .sql(self.dialect)
        )
        con = self.con
        with con.cursor() as cursor, con.transaction():
            catalogs = cursor.execute(cats).fetchall()

        return self._filter_with_like(map(itemgetter(0), catalogs), like)

    def list_databases(
        self, *, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        dbs = (
            sg.select(C.schema_name)
            .from_(sg.table("schemata", db="information_schema"))
            .sql(self.dialect)
        )
        con = self.con
        with con.cursor() as cursor, con.transaction():
            databases = cursor.execute(dbs).fetchall()

        return self._filter_with_like(map(itemgetter(0), databases), like)

    @property
    def current_catalog(self) -> str:
        sql = sg.select(sg.func("current_database")).sql(self.dialect)
        con = self.con
        with con.cursor() as cursor, con.transaction():
            [(db,)] = cursor.execute(sql).fetchall()
        return db

    @property
    def current_database(self) -> str:
        sql = sg.select(sg.func("current_schema")).sql(self.dialect)
        con = self.con
        with con.cursor() as cursor, con.transaction():
            [(schema,)] = cursor.execute(sql).fetchall()
        return schema

    def function(self, name: str, *, database: str | None = None) -> Callable:
        n = ColGen(table="n")
        p = ColGen(table="p")
        f = self.compiler.f

        predicates = [p.proname.eq(sge.convert(name))]

        if database is not None:
            predicates.append(n.nspname.rlike(sge.convert(f"^({database})$")))

        query = (
            sg.select(
                f["pg_catalog.pg_get_function_result"](p.oid).as_("return_type"),
                f.string_to_array(
                    f["pg_catalog.pg_get_function_arguments"](p.oid), ", "
                ).as_("signature"),
            )
            .from_(sg.table("pg_proc", db="pg_catalog").as_("p"))
            .join(
                sg.table("pg_namespace", db="pg_catalog").as_("n"),
                on=n.oid.eq(p.pronamespace),
                join_type="LEFT",
            )
            .where(*predicates)
            .sql(self.dialect)
        )

        def split_name_type(arg: str) -> tuple[str, dt.DataType]:
            name, typ = arg.split(" ", 1)
            return name, self.compiler.type_mapper.from_string(typ)

        con = self.con
        with con.cursor() as cursor, con.transaction():
            rows = cursor.execute(query).fetchall()

        if not rows:
            name = f"{database}.{name}" if database else name
            raise exc.MissingUDFError(name)
        elif len(rows) > 1:
            raise exc.AmbiguousUDFError(name)

        [(raw_return_type, signature)] = rows
        return_type = self.compiler.type_mapper.from_string(raw_return_type)
        signature = list(map(split_name_type, signature))

        # dummy callable
        def fake_func(*args, **kwargs): ...

        fake_func.__name__ = name
        fake_func.__signature__ = inspect.Signature(
            [
                inspect.Parameter(
                    name, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=typ
                )
                for name, typ in signature
            ],
            return_annotation=return_type,
        )
        fake_func.__annotations__ = {"return": return_type, **dict(signature)}
        op = ops.udf.scalar.builtin(fake_func, database=database)
        return op

    def get_schema(
        self,
        name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ):
        # If no database is specified, assume the current database
        dbs = [database or self.current_database]

        # If a database isn't specified, then include temp tables in the
        # returned values
        if database is None and (temp_table_db := self._session_temp_db) is not None:
            dbs.append(temp_table_db)

        type_info = """\
SELECT
  a.attname AS column_name,
  CASE
    WHEN EXISTS(
      SELECT 1
      FROM pg_catalog.pg_type t
      INNER JOIN pg_catalog.pg_enum e
              ON e.enumtypid = t.oid
             AND t.typname = pg_catalog.format_type(a.atttypid, a.atttypmod)
    ) THEN 'enum'
    ELSE pg_catalog.format_type(a.atttypid, a.atttypmod)
  END AS data_type,
  NOT a.attnotnull AS nullable
FROM pg_catalog.pg_attribute a
INNER JOIN pg_catalog.pg_class c
   ON a.attrelid = c.oid
INNER JOIN pg_catalog.pg_namespace n
   ON c.relnamespace = n.oid
WHERE a.attnum > 0
  AND NOT a.attisdropped
  AND n.nspname = ANY(%(dbs)s)
  AND c.relname = %(name)s
ORDER BY a.attnum ASC"""
        type_mapper = self.compiler.type_mapper

        con = self.con
        params = {"dbs": dbs, "name": name}
        with con.cursor() as cursor, con.transaction():
            rows = cursor.execute(type_info, params, prepare=True).fetchall()

        if not rows:
            raise com.TableNotFound(name)

        return sch.Schema(
            {
                col: type_mapper.from_string(typestr, nullable=nullable)
                for col, typestr, nullable in rows
            }
        )

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        name = util.gen_name(f"{self.name}_metadata")

        create_stmt = sge.Create(
            kind="VIEW",
            this=sg.table(name),
            expression=sg.parse_one(query, read=self.dialect),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        ).sql(self.dialect)

        drop_stmt = sge.Drop(kind="VIEW", this=sg.table(name), exists=True).sql(
            self.dialect
        )

        con = self.con
        with con.cursor() as cursor, con.transaction():
            cursor.execute(create_stmt)

        try:
            return self.get_schema(name)
        finally:
            with con.cursor() as cursor, con.transaction():
                cursor.execute(drop_stmt)

    def create_database(
        self, name: str, /, *, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None and catalog != self.current_catalog:
            raise exc.UnsupportedOperationError(
                f"{self.name} does not support creating a database in a different catalog"
            )
        sql = sge.Create(
            kind="SCHEMA", this=sg.table(name, catalog=catalog), exists=force
        ).sql(self.dialect)
        con = self.con
        with con.cursor() as cursor, con.transaction():
            cursor.execute(sql)

    def drop_database(
        self,
        name: str,
        /,
        *,
        catalog: str | None = None,
        force: bool = False,
        cascade: bool = False,
    ) -> None:
        if catalog is not None and catalog != self.current_catalog:
            raise exc.UnsupportedOperationError(
                f"{self.name} does not support dropping a database in a different catalog"
            )

        sql = sge.Drop(
            kind="SCHEMA",
            this=sg.table(name),
            exists=force,
            cascade=cascade,
        ).sql(self.dialect)

        con = self.con
        with con.cursor() as cursor, con.transaction():
            cursor.execute(sql)

    def create_table(
        self,
        name: str,
        /,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Postgres.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists
        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        if not schema:
            schema = table.schema()

        quoted = self.compiler.quoted
        dialect = self.dialect

        table_expr = sg.table(temp_name, db=database, quoted=quoted)
        target = sge.Schema(
            this=table_expr, expressions=schema.to_sqlglot(self.dialect)
        )

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        ).sql(dialect)

        this = sg.table(name, catalog=database, quoted=quoted)
        this_no_catalog = sg.table(name, quoted=quoted)

        con = self.con
        stmts = [create_stmt]

        if query is not None:
            stmts.append(sge.Insert(this=table_expr, expression=query).sql(dialect))

        if overwrite:
            stmts.append(sge.Drop(kind="TABLE", this=this, exists=True).sql(dialect))
            stmts.append(
                f"ALTER TABLE IF EXISTS {table_expr.sql(dialect)} RENAME TO {this_no_catalog.sql(dialect)}"
            )

        with con.cursor() as cursor, con.transaction():
            for stmt in stmts:
                cursor.execute(stmt)

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def drop_table(
        self,
        name: str,
        /,
        *,
        database: str | None = None,
        force: bool = False,
    ) -> None:
        drop_stmt = sg.exp.Drop(
            kind="TABLE",
            this=sg.table(name, db=database, quoted=self.compiler.quoted),
            exists=force,
        ).sql(self.dialect)
        con = self.con
        with con.cursor() as cursor, con.transaction():
            cursor.execute(drop_stmt)

    @contextlib.contextmanager
    def _safe_raw_sql(self, query: str | sg.Expression, **kwargs: Any):
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)

        con = self.con
        with con.cursor() as cursor, con.transaction():
            yield cursor.execute(query, **kwargs)

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)

        con = self.con
        cursor = con.cursor()

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            cursor.close()
            con.rollback()
            raise
        else:
            return cursor

    @util.experimental
    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        import pyarrow as pa

        def _batches(self: Self, *, struct_type: pa.StructType, query: str):
            con = self.con
            # server-side cursors need to be uniquely named
            with (
                con.cursor(name=util.gen_name("postgres_cursor")) as cursor,
                con.transaction(),
            ):
                cur = cursor.execute(query)
                while batch := cur.fetchmany(chunk_size):
                    yield pa.RecordBatch.from_struct_array(
                        pa.array(batch, type=struct_type)
                    )

        self._run_pre_execute_hooks(expr)

        raw_schema = expr.as_table().schema()
        query = self.compile(expr, limit=limit, params=params)
        return pa.RecordBatchReader.from_batches(
            raw_schema.to_pyarrow(),
            _batches(
                self, struct_type=raw_schema.as_struct().to_pyarrow(), query=query
            ),
        )



from xorq.expr.ml.cross_validation import (
    deferred_cross_val_score,
    make_deterministic_sort_key,
)
from xorq.expr.ml.fit_lib import (
    deferred_fit_predict,
    deferred_fit_predict_sklearn,
    deferred_fit_transform,
    deferred_fit_transform_series_sklearn,
    deferred_fit_transform_sklearn,
    deferred_fit_transform_sklearn_struct,
)
from xorq.expr.ml.metrics import (
    deferred_sklearn_metric,
)
from xorq.expr.ml.quickgrove_lib import (
    collect_predicates,  # noqa: F401
    make_quickgrove_udf,
    rewrite_quickgrove_expr,
)
from xorq.expr.ml.split_lib import (
    _calculate_bounds,  # noqa: F401
    calc_split_column,  # noqa: F401
    train_test_splits,
)


also = []
try:
    from xorq.expr.ml.pipeline_lib import (
        FittedPipeline,  # noqa: F401
        Pipeline,  # noqa: F401
        Step,  # noqa: F401
    )

    also = [
        "FittedPipeline",
        "Pipeline",
        "Step",
    ]
except ImportError:
    pass


__all__ = [
    "deferred_cross_val_score",
    "make_deterministic_sort_key",
    "train_test_splits",
    "make_quickgrove_udf",
    "rewrite_quickgrove_expr",
    "deferred_fit_predict",
    "deferred_fit_predict_sklearn",
    "deferred_fit_transform",
    "deferred_fit_transform_sklearn",
    "deferred_fit_transform_sklearn_struct",
    "deferred_fit_transform_series_sklearn",
    "deferred_sklearn_metric",
] + also



import pytest

import xorq.api as xo
from xorq.api import _
from xorq.common.utils.defer_utils import (
    deferred_read_csv,
    deferred_read_parquet,
)
from xorq.expr.relations import into_backend
from xorq.ibis_yaml.compiler import YamlExpressionTranslator


@pytest.fixture(scope="session")
def duckdb_path(tmp_path_factory):
    db_path = tmp_path_factory.mktemp("duckdb") / "test.db"
    return str(db_path)


@pytest.fixture(scope="session")
def prepare_duckdb_con(duckdb_path):
    con = xo.duckdb.connect(duckdb_path)
    con.profile_name = "my_duckdb"  # patch

    con.raw_sql(
        """
        CREATE TABLE IF NOT EXISTS mytable (
            id INT,
            val VARCHAR
        )
        """
    )
    con.raw_sql(
        """
        INSERT INTO mytable
        SELECT i, 'val' || i::VARCHAR
        FROM range(1, 6) t(i)
        """
    )
    return con


def test_duckdb_database_table_roundtrip(prepare_duckdb_con):
    con = prepare_duckdb_con
    profiles = {con._profile.hash_name: con}
    table_expr = con.table("mytable")

    expr1 = table_expr.mutate(new_val=(table_expr.val + "_extra"))
    yaml_dict = YamlExpressionTranslator.to_yaml(expr1, profiles)
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    df_original = expr1.execute()
    df_roundtrip = roundtrip_expr.execute()

    assert df_original.equals(df_roundtrip), "Roundtrip expression data differs!"


@pytest.mark.xfail(reason="MemTable is not serializable")
def test_memtable(builds_dir):
    table = xo.memtable([(i, "val") for i in range(10)], columns=["key1", "val"])
    backend = table._find_backend()
    expr = table.mutate(new_val=2 * xo._.val)

    profiles = {backend._profile.hash_name: backend}

    yaml_dict = YamlExpressionTranslator.to_yaml(expr, profiles, builds_dir)
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict)

    expr.equals(roundtrip_expr)

    assert expr.execute().equals(roundtrip_expr.execute())


def test_into_backend():
    parquet_path = xo.config.options.pins.get_path("awards_players")
    backend = xo.duckdb.connect()
    table = deferred_read_parquet(parquet_path, backend, table_name="award_players")
    expr = table.mutate(new_id=2 * xo._.playerID)

    con2 = xo.connect()
    con3 = xo.connect()

    expr = into_backend(expr, con2, "ls_mem").mutate(x=4 * xo._.new_id)
    expr = into_backend(expr, con3, "df_mem")

    profiles = {
        backend._profile.hash_name: backend,
        con2._profile.hash_name: con2,
        con3._profile.hash_name: con3,
    }

    yaml_dict = YamlExpressionTranslator.to_yaml(expr, profiles)
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    assert xo.execute(expr).equals(xo.execute(roundtrip_expr))


@pytest.mark.xfail(reason="MemTable is not serializable")
def test_memtable_cache():
    table = xo.memtable([(i, "val") for i in range(10)], columns=["key1", "val"])
    expr = table.mutate(new_val=2 * xo._.val).cache()
    yaml_dict = YamlExpressionTranslator.to_yaml(expr)
    roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict)

    assert xo.execute(expr).equals(xo.execute(roundtrip_expr))


def test_deferred_read_csv(csv_dir):
    csv_name = "astronauts"
    csv_path = csv_dir / f"{csv_name}.csv"

    pd_con = xo.pandas.connect()
    expr = deferred_read_csv(con=pd_con, path=csv_path, table_name=csv_name).filter(
        _.id > 6
    )

    profiles = {pd_con._profile.hash_name: pd_con}
    with pytest.warns(UserWarning):
        yaml_dict = YamlExpressionTranslator.to_yaml(expr, profiles)
        roundtrip_expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles)

    assert xo.execute(expr).equals(xo.execute(roundtrip_expr))



from __future__ import annotations

import numpy as np
import pytest

import xorq.api as xo
from xorq.api import _


def test_predicate_collection(prediction_expr):
    """Verifies that a single filter predicate is correctly collected from an expression"""
    expr = prediction_expr.filter(_.carat < 1.0)
    predicates = xo.expr.ml.collect_predicates(expr.op())
    assert len(predicates) == 1
    assert predicates[0]["column"] == "carat"
    assert predicates[0]["op"] == "Less"
    assert predicates[0]["value"] == 1.0


def test_predicate_collection_on_udf_column_raises(prediction_expr):
    """Confirms that filter predicates on UDF columns are not collected"""
    expr = prediction_expr.filter(_.pred < 1.0)
    with pytest.raises(ValueError, match="Unsupported predicate on UDF column: pred"):
        xo.expr.ml.collect_predicates(expr.op())


def test_multiple_predicates(prediction_expr):
    """Checks that multiple filter predicates are correctly collected with their operators"""
    expr = prediction_expr.filter([_.carat < 1.0, _.depth > 60, _.table >= 55])
    predicates = xo.expr.ml.collect_predicates(expr.op())
    assert len(predicates) == 3
    assert {p["op"] for p in predicates} == {"Less", "Greater", "GreaterEqual"}


def test_model_pruning(prediction_expr):
    """Ensures pruned model predictions match original model predictions"""
    original = prediction_expr.filter(_.carat < 1.0).execute()
    optimized = xo.expr.ml.rewrite_quickgrove_expr(
        prediction_expr.filter(_.carat < 1.0)
    ).execute()
    np.testing.assert_array_almost_equal(
        original["pred"].values, optimized["pred"].values, decimal=3
    )


@pytest.mark.xfail
def test_pruned_model_exists_in_repr(prediction_expr):
    """Checks that pruned models are correctly marked in their string representation"""
    original = prediction_expr.filter(_.carat < 1.0)
    optimized = xo.expr.ml.rewrite_quickgrove_expr(
        prediction_expr.filter(_.carat < 1.0)
    )
    assert "_pruned" not in repr(original)
    assert "_pruned" in repr(optimized)


def test_mixed_pruned_model_types(mixed_prediction_expr):
    """Verifies that mixed model types maintain prediction accuracy after pruning"""
    original = mixed_prediction_expr.filter(_.carat < 1.0).execute()
    optimized = xo.expr.ml.rewrite_quickgrove_expr(
        mixed_prediction_expr.filter(_.carat < 1.0)
    ).execute()
    np.testing.assert_array_almost_equal(
        original["pred"].values, optimized["pred"].values, decimal=3
    )


def test_prune_if_predicates_are_on_udf(mixed_model_path, mixed_feature_table):
    """Verifies that models aren't pruned when filters are on UDF columns"""
    predict_udf = xo.expr.ml.make_quickgrove_udf(mixed_model_path)
    t = mixed_feature_table.mutate(pred=predict_udf.on_expr).filter(_.pred < 1)
    with pytest.raises(ValueError):
        xo.expr.ml.rewrite_quickgrove_expr(t)


def test_prune_if_predicates_are_not_features(mixed_model_path, mixed_feature_table):
    """Confirms that models aren't pruned when filters are on non-feature columns"""
    predict_udf = xo.expr.ml.make_quickgrove_udf(mixed_model_path)
    t = mixed_feature_table.mutate(pred=predict_udf.on_expr).filter(_.target >= 1)
    optimized = xo.expr.ml.rewrite_quickgrove_expr(t)
    assert "pruned" not in repr(optimized)


def test_mixed_prune_greather_than(mixed_prediction_expr):
    """Tests model pruning with greater than comparisons"""
    original = mixed_prediction_expr.filter(_.carat > 1.0).execute()
    optimized = xo.expr.ml.rewrite_quickgrove_expr(
        mixed_prediction_expr.filter(_.carat > 1.0)
    ).execute()
    np.testing.assert_array_almost_equal(
        original["pred"].values, optimized["pred"].values, decimal=3
    )


def test_mixed_prune_greather_than_or_equal(mixed_prediction_expr):
    """Tests model pruning with greater than or equal comparisons"""
    original = mixed_prediction_expr.filter(_.carat >= 1.0).execute()
    optimized = xo.expr.ml.rewrite_quickgrove_expr(
        mixed_prediction_expr.filter(_.carat >= 1.0)
    ).execute()
    np.testing.assert_array_almost_equal(
        original["pred"].values, optimized["pred"].values, decimal=3
    )



import operator

import toolz

import xorq.api as xo
import xorq.expr.relations as rel
import xorq.vendor.ibis.expr.operations as ops
from xorq.caching import SourceCache
from xorq.common.utils.graph_utils import (
    find_all_sources,
    walk_nodes,
)


node_types = (
    ops.DatabaseTable,
    ops.SQLQueryResult,
    rel.CachedNode,
    rel.Read,
    rel.RemoteTable,
    # ExprScalarUDF has an expr we need to get to
    # FlightOperator has a dynamically generated connection: it should be passed a Profile instead
)


def make_expr():
    cons = (con0, con1, con2, con3) = (
        xo.connect(),
        xo.connect(),
        xo.duckdb.connect(),
        xo.connect(),
    )

    read_node0 = xo.examples.awards_players.fetch(con0)
    remote_node0 = read_node0.into_backend(con1)
    cached_node0 = remote_node0.cache(SourceCache.from_kwargs(source=con1))
    read_node1 = xo.examples.batting.fetch(con2)
    remote_node1 = read_node1.into_backend(con1)
    remote_node2 = cached_node0.join(
        remote_node1, predicates=["playerID", "yearID", "lgID"]
    ).into_backend(con3)
    cached_node1 = remote_node2.cache()
    expr = cached_node2 = cached_node1[lambda t: t.G == 1].cache()
    nodes = {
        rel.CachedNode: (
            cached_node0.op(),
            cached_node1.op(),
            cached_node2.op(),
        ),
        rel.Read: (
            read_node0.op(),
            read_node1.op(),
        ),
        rel.RemoteTable: (
            remote_node0.op(),
            remote_node1.op(),
            remote_node2.op(),
        ),
    }
    return (cons, nodes, expr)


def test_walk_nodes():
    (_, nodes, expr) = make_expr()
    node_types = tuple(nodes)
    walked_nodes = walk_nodes(node_types, expr)
    expected = sorted(
        ((k, set(v)) for k, v in nodes.items()),
        key=toolz.compose(operator.attrgetter("__name__"), operator.itemgetter(0)),
    )
    actual = sorted(
        ((k, set(v)) for k, v in toolz.groupby(type, walked_nodes).items()),
        key=toolz.compose(operator.attrgetter("__name__"), operator.itemgetter(0)),
    )
    assert actual == expected


def test_find_all_sources():
    (created_sources, _, expr) = make_expr()
    found_sources = find_all_sources(expr)
    actual = {con._profile for con in created_sources}
    expected = {con._profile for con in found_sources}
    assert actual == expected


def test_replace_computed_kwargs_expr(parquet_dir):
    from sklearn.linear_model import LinearRegression

    import xorq.expr.datatypes as dt
    import xorq.expr.selectors as s
    from xorq.common.utils.graph_utils import (
        walk_nodes,
    )
    from xorq.expr.relations import (
        Tag,
    )
    from xorq.ml import (
        deferred_fit_predict_sklearn,
    )

    deferred_linear_regression = deferred_fit_predict_sklearn(
        cls=LinearRegression, return_type=dt.float64
    )

    t = xo.deferred_read_parquet(parquet_dir / "diamonds.parquet", xo.connect())
    train_table, test_table = (
        el.tag(tag)
        for el, tag in zip(
            xo.train_test_splits(
                t, unique_key=tuple(t.columns), test_sizes=0.5, random_seed=42
            ),
            ("train", "test"),
        )
    )
    target = "price"
    features = tuple(c for c in t.select(s.numeric()).columns if c != target)
    predict_expr_udf = deferred_linear_regression(
        train_table, target, features
    ).deferred_other
    predicted = test_table.mutate(predict_expr_udf.on_expr(test_table))
    assert walk_nodes(Tag, predicted)
    removed = xo.expr.api._remove_tag_nodes(predicted)
    assert not walk_nodes(Tag, removed)
    assert not walk_nodes(Tag, predicted.ls.untagged)



import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)


@pytest.mark.parametrize(
    "get_expr",
    [
        lambda t: t,
        lambda t: t.group_by("playerID").agg(t.stint.max().name("n-stints")),
    ],
)
def test_register_with_different_name_and_cache(csv_dir, get_expr):
    batting_path = csv_dir.joinpath("batting.csv")
    table_name = "batting"

    datafusion_con = xo.datafusion.connect()
    xorq_table_name = f"{datafusion_con.name}_{table_name}"
    t = datafusion_con.read_csv(
        batting_path, table_name=table_name, schema_infer_max_records=50_000
    )
    expr = t.pipe(get_expr).cache()

    assert table_name != xorq_table_name
    assert expr.execute() is not None


def test_datafusion_snapshot(con_snapshot):
    con_snapshot(xo.datafusion.connect())


def test_cross_source_snapshot(con_cross_source_snapshot):
    con_cross_source_snapshot(xo.datafusion.connect())


@pytest.mark.parametrize(
    "cls",
    [ParquetSnapshotCache, ParquetCache, SourceSnapshotCache, SourceCache],
)
def test_cache_find_backend(cls, parquet_dir, con_cache_find_backend):
    con_cache_find_backend(cls, xo.datafusion.connect())



import pyarrow as pa
import pytest

from xorq.backends.pyiceberg.tests.conftest import QUOTES_TABLE_NAME
from xorq.vendor.ibis import Schema


def test_create_table_from_expr(iceberg_con, quotes_table):
    t = quotes_table.select("timestamp", "bid", "ask").limit(10).as_table()

    actual_t = iceberg_con.create_table("selected_quotes", t)

    assert not actual_t.execute().empty


def identity(val):
    return val


def execute(val):
    val = val.execute()
    val["timestamp"] = val["timestamp"].astype("datetime64[us]")
    return val


@pytest.mark.parametrize(
    "fun",
    [
        pytest.param(execute, id="pandas"),
        pytest.param(identity, id="xorq"),
    ],
)
def test_append(iceberg_con, trades_df, fun):
    table_name = "trades"

    t = iceberg_con.create_table(table_name, trades_df, overwrite=True)
    assert table_name in iceberg_con.list_tables()
    expected = t.execute()

    iceberg_con.insert(table_name, fun(t))
    actual = t.execute()
    assert len(actual) == len(expected) * 2


def test_create_empty_raises(iceberg_con):
    with pytest.raises(NotImplementedError):
        schema = Schema.from_pyarrow(
            pa.schema(
                [
                    pa.field("question", pa.string()),
                    pa.field("product_description", pa.string()),
                    pa.field("image_url", pa.string()),
                    pa.field("label", pa.uint8(), nullable=True),
                ]
            )
        )
        iceberg_con.create_table("empty", schema=schema)


def test_read_record_batches(iceberg_con, quotes_table):
    table_name = "quotes_records"
    t = iceberg_con.read_record_batches(
        quotes_table.to_pyarrow_batches(), table_name=table_name
    )
    assert table_name in iceberg_con.list_tables()
    assert not t.execute().empty


def test_get_schema(iceberg_con):
    iceberg_con.get_schema(QUOTES_TABLE_NAME)



import pandas as pd
import pytest

import xorq.api as xo


@pytest.fixture(scope="session")
def sqlite_con():
    return xo.sqlite.connect()


@pytest.fixture(scope="session")
def astronauts_parquet_path(parquet_dir):
    return parquet_dir / "astronauts.parquet"


@pytest.fixture(scope="session")
def astronauts_csv_path(csv_dir):
    return csv_dir / "astronauts.csv"


@pytest.fixture(scope="session")
def persistent_sqlite_con(tmp_path_factory):
    path = tmp_path_factory.mktemp("database").joinpath("lite.db").resolve()
    return xo.sqlite.connect(str(path))


@pytest.fixture(scope="session")
def df():
    return pd.DataFrame(
        list(
            zip(
                *[
                    [0, 1.5, 2.3, 3, 4, 5, 6],
                    [7, 4, 3, 8, 9, 1, 6],
                    ["A", "A", "A", "A", "B", "B", "B"],
                ]
            )
        ),
        columns=["a", "b", "c"],
    )



import pytest

import xorq.api as xo
from xorq.api import _
from xorq.ibis_yaml.compiler import build_expr


@pytest.mark.benchmark
def test_baseball_stats_compilation(builds_dir):
    pg = xo.postgres.connect_env()

    batting_old = (
        pg.table("batting")
        .filter(_.yearID < 2000)
        .filter(_.yearID >= 1990)
        .mutate(batting_avg=_.H / _.AB, era=_.yearID.cast("int32"))
    )

    batting_recent = (
        pg.table("batting")
        .filter(_.yearID >= 2000)
        .mutate(batting_avg=_.H / _.AB, era=_.yearID.cast("int32"))
    )

    all_batting = batting_old.union(batting_recent)

    league_averages = all_batting.group_by(["yearID", "lgID"]).aggregate(
        league_avg=_.batting_avg.mean(), league_HR_avg=_.HR.mean()
    )

    normalized_stats = all_batting.join(league_averages, ["yearID", "lgID"]).mutate(
        avg_vs_league=_.batting_avg - _.league_avg, HR_vs_league=_.HR - _.league_HR_avg
    )

    prev_year_stats = (
        normalized_stats.order_by(["playerID", "yearID"])
        .group_by("playerID")
        .mutate(
            prev_year_avg=_.batting_avg.lag(1).over(order_by=_.yearID),
            prev_year_HR=_.HR.lag(1).over(order_by=_.yearID),
            three_year_avg=_.batting_avg.mean().over(
                order_by=_.yearID, window=xo.window(preceding=2, following=0)
            ),
        )
    )

    build_path = build_expr(prev_year_stats, builds_dir=builds_dir)
    expr_hash = build_path.name
    assert isinstance(expr_hash, str)
    assert len(expr_hash) > 0



from __future__ import annotations

import contextlib
import time
from pathlib import Path
from typing import TYPE_CHECKING

import pytest


if TYPE_CHECKING:
    pass

# ── Constants ────────────────────────────────────────────────────────────────
GIZMOSQL_PORT = 31337
GIZMOSQL_IMAGE = "gizmodata/gizmosql:latest"
GIZMOSQL_USERNAME = "ibis"
GIZMOSQL_PASSWORD = "ibis_password"
CONTAINER_NAME = "xorq-gizmosql-test"

ROOT_DIR = Path(__file__).resolve().parents[5]  # xorq repo root
DATA_DIR = ROOT_DIR / "ci" / "ibis-testing-data"

PARQUET_TABLES = (
    "functional_alltypes",
    "diamonds",
    "batting",
    "awards_players",
    "astronauts",
)


# ── Docker container management ──────────────────────────────────────────────
def _port_is_listening(port: int, host: str = "localhost") -> bool:
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        return s.connect_ex((host, port)) == 0


def _wait_for_container_log(
    container,
    timeout=60,
    poll_interval=1,
    ready_message="GizmoSQL server - started",
):
    start_time = time.time()
    while time.time() - start_time < timeout:
        logs = container.logs().decode("utf-8")
        if ready_message in logs:
            return True
        time.sleep(poll_interval)
    raise TimeoutError(f"Container did not show '{ready_message}' within {timeout}s.")


@pytest.fixture(scope="session")
def gizmosql_server():
    """Start the GizmoSQL Docker container for testing.

    If a server is already listening on GIZMOSQL_PORT, skip Docker management
    and use that server instead.
    """
    if _port_is_listening(GIZMOSQL_PORT):
        yield None
        return

    docker = pytest.importorskip("docker")
    client = docker.from_env()
    parquet_dir = str(DATA_DIR / "parquet")

    try:
        container = client.containers.get(CONTAINER_NAME)
        if container.status == "running":
            yield container
            return
        container.remove(force=True)
    except docker.errors.NotFound:
        pass

    container = client.containers.run(
        image=GIZMOSQL_IMAGE,
        name=CONTAINER_NAME,
        detach=True,
        remove=True,
        tty=True,
        init=True,
        ports={f"{GIZMOSQL_PORT}/tcp": GIZMOSQL_PORT},
        volumes={parquet_dir: {"bind": "/data/parquet", "mode": "ro"}},
        environment={
            "GIZMOSQL_USERNAME": GIZMOSQL_USERNAME,
            "GIZMOSQL_PASSWORD": GIZMOSQL_PASSWORD,
            "TLS_ENABLED": "1",
            "PRINT_QUERIES": "0",
            "DATABASE_FILENAME": ":memory:",
        },
        stdout=True,
        stderr=True,
    )

    _wait_for_container_log(container)
    yield container
    container.stop()


@pytest.fixture(scope="session")
def con(gizmosql_server):
    """GizmoSQL connection with test data loaded."""
    import xorq.api as xo

    conn = xo.gizmosql.connect(
        host="localhost",
        user=GIZMOSQL_USERNAME,
        password=GIZMOSQL_PASSWORD,
        port=GIZMOSQL_PORT,
        use_encryption=True,
        disable_certificate_verification=True,
    )

    # Load standard test tables from parquet
    parquet_dir = DATA_DIR / "parquet"
    for table_name in PARQUET_TABLES:
        parquet_path = parquet_dir / f"{table_name}.parquet"
        if parquet_path.exists():
            conn.read_parquet(parquet_path, table_name=table_name)

    return conn


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def awards_players(con):
    return con.table("awards_players")


@pytest.fixture(scope="session")
def batting(con):
    return con.table("batting")


@pytest.fixture
def temp_table(con):
    from xorq.vendor.ibis import util

    name = util.gen_name("temp_table")
    yield name
    with contextlib.suppress(Exception):
        with con._safe_raw_sql(f'DROP TABLE IF EXISTS "{name}"'):
            pass



from __future__ import annotations

import datetime

from xorq.vendor.ibis.formats.pandas import PandasData


class TrinoPandasData(PandasData):
    @classmethod
    def convert_Interval(cls, s, dtype, pandas_dtype):
        def parse_trino_timedelta(value):
            # format is 'days hour:minute:second.millisecond'
            days, rest = value.split(" ", 1)
            hms, millis = rest.split(".", 1)
            hours, minutes, seconds = hms.split(":")
            return datetime.timedelta(
                days=int(days),
                hours=int(hours),
                minutes=int(minutes),
                seconds=int(seconds),
                milliseconds=int(millis),
            )

        return s.map(parse_trino_timedelta, na_action="ignore")



import ibis
import ibis.expr.operations as ops

from xorq.common.utils.ibis_utils import from_ibis


def test_scalar_subquery(t):
    expr = ops.ScalarSubquery(t.c.mean().as_table()).to_expr()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_exists_subquery():
    t1 = ibis.table({"a": "int", "b": "string"}, name="t1")
    t2 = ibis.table({"a": "int", "c": "float"}, name="t2")

    filtered = t2.filter(t2.a == t1.a)
    expr = ops.ExistsSubquery(filtered).to_expr()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_in_subquery():
    t1 = ibis.table({"a": "int", "b": "string"}, name="t1")
    t2 = ibis.table({"a": "int", "c": "float"}, name="t2")

    expr = ops.InSubquery(t1.select("a"), t2.a).to_expr()

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None



from __future__ import annotations

import functools
from collections.abc import Sequence
from typing import TYPE_CHECKING, Literal

from public import public

import xorq.vendor.ibis.expr.operations as ops
from xorq.common.exceptions import XorqTypeError
from xorq.vendor import ibis
from xorq.vendor.ibis.expr.types.core import _binop
from xorq.vendor.ibis.expr.types.generic import Column, Scalar, Value
from xorq.vendor.ibis.util import deprecated


if TYPE_CHECKING:
    from collections.abc import Iterable

    import xorq.vendor.ibis.expr.types as ir


@public
class NumericValue(Value):
    def negate(self) -> NumericValue:
        """Negate a numeric expression.

        Returns
        -------
        NumericValue
            A numeric value expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.negate()
        ┏━━━━━━━━━━━━━━━━┓
        ┃ Negate(values) ┃
        ┡━━━━━━━━━━━━━━━━┩
        │ int64          │
        ├────────────────┤
        │              1 │
        │              0 │
        │             -1 │
        └────────────────┘
        """
        return ops.Negate(self).to_expr()

    def __neg__(self) -> NumericValue:
        """Negate `self`.

        Returns
        -------
        NumericValue
            `self` negated
        """
        return self.negate()

    def round(self, digits: int | IntegerValue | None = None) -> NumericValue:
        """Round values to an indicated number of decimal places.

        Parameters
        ----------
        digits
            The number of digits to round to.

            Here's how the `digits` parameter affects the expression output
            type:

            - `digits` is `False`-y; `self.type()` is `decimal` → `decimal`
            -   `digits` is nonzero; `self.type()` is `decimal` → `decimal`
            - `digits` is `False`-y; `self.type()` is Floating  → `int64`
            -   `digits` is nonzero; `self.type()` is Floating  → `float64`

        Returns
        -------
        NumericValue
            The rounded expression

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1.22, 1.64, 2.15, 2.54]})
        >>> t
        ┏━━━━━━━━━┓
        ┃ values  ┃
        ┡━━━━━━━━━┩
        │ float64 │
        ├─────────┤
        │    1.22 │
        │    1.64 │
        │    2.15 │
        │    2.54 │
        └─────────┘
        >>> t.values.round()
        ┏━━━━━━━━━━━━━━━┓
        ┃ Round(values) ┃
        ┡━━━━━━━━━━━━━━━┩
        │ int64         │
        ├───────────────┤
        │             1 │
        │             2 │
        │             2 │
        │             3 │
        └───────────────┘
        >>> t.values.round(1)
        ┏━━━━━━━━━━━━━━━━━━┓
        ┃ Round(values, 1) ┃
        ┡━━━━━━━━━━━━━━━━━━┩
        │ float64          │
        ├──────────────────┤
        │              1.2 │
        │              1.6 │
        │              2.2 │
        │              2.5 │
        └──────────────────┘
        """
        return ops.Round(self, digits).to_expr()

    def log(self, base: NumericValue | None = None) -> NumericValue:
        r"""Compute $\log_{\texttt{base}}\left(\texttt{self}\right)$.

        Parameters
        ----------
        base
            The base of the logarithm. If `None`, base `e` is used.

        Returns
        -------
        NumericValue
            Logarithm of `arg` with base `base`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> from math import e
        >>> t = ibis.memtable({"values": [e, e**2, e**3]})
        >>> t.values.log()
        ┏━━━━━━━━━━━━━┓
        ┃ Log(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │         1.0 │
        │         2.0 │
        │         3.0 │
        └─────────────┘


        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [10, 100, 1000]})
        >>> t.values.log(10)
        ┏━━━━━━━━━━━━━━━━━┓
        ┃ Log(values, 10) ┃
        ┡━━━━━━━━━━━━━━━━━┩
        │ float64         │
        ├─────────────────┤
        │             1.0 │
        │             2.0 │
        │             3.0 │
        └─────────────────┘
        """
        return ops.Log(self, base).to_expr()

    def clip(
        self,
        lower: NumericValue | None = None,
        upper: NumericValue | None = None,
    ) -> NumericValue:
        """Trim values outside of `lower` and `upper` bounds.

        `NULL` values are preserved and are not replaced with bounds.

        Parameters
        ----------
        lower
            Lower bound
        upper
            Upper bound

        Returns
        -------
        NumericValue
            Clipped input

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable(
        ...     {"values": [None, 2, 3, None, 5, None, None, 8]},
        ...     schema=dict(values="int"),
        ... )
        >>> t.values.clip(lower=3, upper=6)
        ┏━━━━━━━━━━━━━━━━━━━━┓
        ┃ Clip(values, 3, 6) ┃
        ┡━━━━━━━━━━━━━━━━━━━━┩
        │ int64              │
        ├────────────────────┤
        │               NULL │
        │                  3 │
        │                  3 │
        │               NULL │
        │                  5 │
        │               NULL │
        │               NULL │
        │                  6 │
        └────────────────────┘
        """
        if lower is None and upper is None:
            raise ValueError("at least one of lower and upper must be provided")

        return ops.Clip(self, lower, upper).to_expr()

    def abs(self) -> NumericValue:
        """Return the absolute value of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 2, -3, 4]})
        >>> t.values.abs()
        ┏━━━━━━━━━━━━━┓
        ┃ Abs(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ int64       │
        ├─────────────┤
        │           1 │
        │           2 │
        │           3 │
        │           4 │
        └─────────────┘
        """
        return ops.Abs(self).to_expr()

    def ceil(self) -> DecimalValue | IntegerValue:
        """Return the ceiling of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 1.1, 2, 2.1, 3.3]})
        >>> t.values.ceil()
        ┏━━━━━━━━━━━━━━┓
        ┃ Ceil(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ int64        │
        ├──────────────┤
        │            1 │
        │            2 │
        │            2 │
        │            3 │
        │            4 │
        └──────────────┘
        """
        return ops.Ceil(self).to_expr()

    def degrees(self) -> NumericValue:
        """Compute the degrees of `self` radians.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> from math import pi
        >>> t = ibis.memtable({"values": [0, pi / 2, pi, 3 * pi / 2, 2 * pi]})
        >>> t.values.degrees()
        ┏━━━━━━━━━━━━━━━━━┓
        ┃ Degrees(values) ┃
        ┡━━━━━━━━━━━━━━━━━┩
        │ float64         │
        ├─────────────────┤
        │             0.0 │
        │            90.0 │
        │           180.0 │
        │           270.0 │
        │           360.0 │
        └─────────────────┘
        """
        return ops.Degrees(self).to_expr()

    rad2deg = degrees

    def exp(self) -> NumericValue:
        r"""Compute $e^\texttt{self}$.

        Returns
        -------
        NumericValue
            $e^\texttt{self}$

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": range(4)})
        >>> t.values.exp()
        ┏━━━━━━━━━━━━━┓
        ┃ Exp(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │    1.000000 │
        │    2.718282 │
        │    7.389056 │
        │   20.085537 │
        └─────────────┘
        """
        return ops.Exp(self).to_expr()

    def floor(self) -> DecimalValue | IntegerValue:
        """Return the floor of an expression.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 1.1, 2, 2.1, 3.3]})
        >>> t.values.floor()
        ┏━━━━━━━━━━━━━━━┓
        ┃ Floor(values) ┃
        ┡━━━━━━━━━━━━━━━┩
        │ int64         │
        ├───────────────┤
        │             1 │
        │             1 │
        │             2 │
        │             2 │
        │             3 │
        └───────────────┘

        """
        return ops.Floor(self).to_expr()

    def log2(self) -> NumericValue:
        r"""Compute $\log_{2}\left(\texttt{self}\right)$.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 2, 4, 8]})
        >>> t.values.log2()
        ┏━━━━━━━━━━━━━━┓
        ┃ Log2(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ float64      │
        ├──────────────┤
        │          0.0 │
        │          1.0 │
        │          2.0 │
        │          3.0 │
        └──────────────┘
        """
        return ops.Log2(self).to_expr()

    def log10(self) -> NumericValue:
        r"""Compute $\log_{10}\left(\texttt{self}\right)$.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 10, 100]})
        >>> t.values.log10()
        ┏━━━━━━━━━━━━━━━┓
        ┃ Log10(values) ┃
        ┡━━━━━━━━━━━━━━━┩
        │ float64       │
        ├───────────────┤
        │           0.0 │
        │           1.0 │
        │           2.0 │
        └───────────────┘
        """
        return ops.Log10(self).to_expr()

    def ln(self) -> NumericValue:
        r"""Compute $\ln\left(\texttt{self}\right)$.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 2.718281828, 3]})
        >>> t.values.ln()
        ┏━━━━━━━━━━━━┓
        ┃ Ln(values) ┃
        ┡━━━━━━━━━━━━┩
        │ float64    │
        ├────────────┤
        │   0.000000 │
        │   1.000000 │
        │   1.098612 │
        └────────────┘
        """
        return ops.Ln(self).to_expr()

    def radians(self) -> NumericValue:
        """Compute radians from `self` degrees.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [0, 90, 180, 270, 360]})
        >>> t.values.radians()
        ┏━━━━━━━━━━━━━━━━━┓
        ┃ Radians(values) ┃
        ┡━━━━━━━━━━━━━━━━━┩
        │ float64         │
        ├─────────────────┤
        │        0.000000 │
        │        1.570796 │
        │        3.141593 │
        │        4.712389 │
        │        6.283185 │
        └─────────────────┘
        """
        return ops.Radians(self).to_expr()

    deg2rad = radians

    def sign(self) -> NumericValue:
        """Return the sign of the input.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 2, -3, 4]})
        >>> t.values.sign()
        ┏━━━━━━━━━━━━━━┓
        ┃ Sign(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ int64        │
        ├──────────────┤
        │           -1 │
        │            1 │
        │           -1 │
        │            1 │
        └──────────────┘
        """
        return ops.Sign(self).to_expr()

    def sqrt(self) -> NumericValue:
        """Compute the square root of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [1, 4, 9, 16]})
        >>> t.values.sqrt()
        ┏━━━━━━━━━━━━━━┓
        ┃ Sqrt(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ float64      │
        ├──────────────┤
        │          1.0 │
        │          2.0 │
        │          3.0 │
        │          4.0 │
        └──────────────┘
        """
        return ops.Sqrt(self).to_expr()

    def acos(self) -> NumericValue:
        """Compute the arc cosine of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.acos()
        ┏━━━━━━━━━━━━━━┓
        ┃ Acos(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ float64      │
        ├──────────────┤
        │     3.141593 │
        │     1.570796 │
        │     0.000000 │
        └──────────────┘

        """
        return ops.Acos(self).to_expr()

    def asin(self) -> NumericValue:
        """Compute the arc sine of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.asin()
        ┏━━━━━━━━━━━━━━┓
        ┃ Asin(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ float64      │
        ├──────────────┤
        │    -1.570796 │
        │     0.000000 │
        │     1.570796 │
        └──────────────┘
        """
        return ops.Asin(self).to_expr()

    def atan(self) -> NumericValue:
        """Compute the arc tangent of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.atan()
        ┏━━━━━━━━━━━━━━┓
        ┃ Atan(values) ┃
        ┡━━━━━━━━━━━━━━┩
        │ float64      │
        ├──────────────┤
        │    -0.785398 │
        │     0.000000 │
        │     0.785398 │
        └──────────────┘
        """
        return ops.Atan(self).to_expr()

    def atan2(self, other: NumericValue) -> NumericValue:
        """Compute the two-argument version of arc tangent.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.atan2(0)
        ┏━━━━━━━━━━━━━━━━━━┓
        ┃ Atan2(values, 0) ┃
        ┡━━━━━━━━━━━━━━━━━━┩
        │ float64          │
        ├──────────────────┤
        │        -1.570796 │
        │         0.000000 │
        │         1.570796 │
        └──────────────────┘
        """
        return ops.Atan2(self, other).to_expr()

    def cos(self) -> NumericValue:
        """Compute the cosine of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.cos()
        ┏━━━━━━━━━━━━━┓
        ┃ Cos(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │    0.540302 │
        │    1.000000 │
        │    0.540302 │
        └─────────────┘
        """
        return ops.Cos(self).to_expr()

    def cot(self) -> NumericValue:
        """Compute the cotangent of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, -2, 3]})
        >>> t.values.cot()
        ┏━━━━━━━━━━━━━┓
        ┃ Cot(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │   -0.642093 │
        │    0.457658 │
        │   -7.015253 │
        └─────────────┘
        """
        return ops.Cot(self).to_expr()

    def sin(self) -> NumericValue:
        """Compute the sine of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.sin()
        ┏━━━━━━━━━━━━━┓
        ┃ Sin(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │   -0.841471 │
        │    0.000000 │
        │    0.841471 │
        └─────────────┘
        """
        return ops.Sin(self).to_expr()

    def tan(self) -> NumericValue:
        """Compute the tangent of `self`.

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"values": [-1, 0, 1]})
        >>> t.values.tan()
        ┏━━━━━━━━━━━━━┓
        ┃ Tan(values) ┃
        ┡━━━━━━━━━━━━━┩
        │ float64     │
        ├─────────────┤
        │   -1.557408 │
        │    0.000000 │
        │    1.557408 │
        └─────────────┘
        """
        return ops.Tan(self).to_expr()

    def __add__(self, other: NumericValue) -> NumericValue:
        """Add `self` with `other`."""
        return _binop(ops.Add, self, other)

    add = radd = __radd__ = __add__

    def __sub__(self, other: NumericValue) -> NumericValue:
        """Subtract `other` from `self`."""
        return _binop(ops.Subtract, self, other)

    sub = __sub__

    def __rsub__(self, other: NumericValue) -> NumericValue:
        """Subtract `self` from `other`."""
        return _binop(ops.Subtract, other, self)

    rsub = __rsub__

    def __mul__(self, other: NumericValue) -> NumericValue:
        """Multiply `self` and `other`."""
        return _binop(ops.Multiply, self, other)

    mul = rmul = __rmul__ = __mul__

    def __truediv__(self, other):
        """Divide `self` by `other`."""
        return _binop(ops.Divide, self, other)

    div = __div__ = __truediv__

    def __rtruediv__(self, other: NumericValue) -> NumericValue:
        """Divide `other` by `self`."""
        return _binop(ops.Divide, other, self)

    rdiv = __rdiv__ = __rtruediv__

    def __floordiv__(
        self,
        other: NumericValue,
    ) -> NumericValue:
        """Floor divide `self` by `other`."""
        return _binop(ops.FloorDivide, self, other)

    floordiv = __floordiv__

    def __rfloordiv__(
        self,
        other: NumericValue,
    ) -> NumericValue:
        """Floor divide `other` by `self`."""
        return _binop(ops.FloorDivide, other, self)

    rfloordiv = __rfloordiv__

    def __pow__(self, other: NumericValue) -> NumericValue:
        """Raise `self` to the `other`th power."""
        return _binop(ops.Power, self, other)

    pow = __pow__

    def __rpow__(self, other: NumericValue) -> NumericValue:
        """Raise `other` to the `self`th power."""
        return _binop(ops.Power, other, self)

    rpow = __rpow__

    def __mod__(self, other: NumericValue) -> NumericValue:
        """Compute `self` modulo `other`."""
        return _binop(ops.Modulus, self, other)

    mod = __mod__

    def __rmod__(self, other: NumericValue) -> NumericValue:
        """Compute `other` modulo `self`."""

        return _binop(ops.Modulus, other, self)

    rmod = __rmod__

    def point(self, right: int | float | NumericValue) -> ir.PointValue:
        """Return a point constructed from the coordinate values.

        Constant coordinates result in construction of a `POINT` literal or
        column.

        Parameters
        ----------
        right
            Y coordinate

        Returns
        -------
        PointValue
            Points

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.zones.fetch()
        >>> t.x_cent.point(t.y_cent)
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ GeoPoint(x_cent, y_cent)         ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ point:geometry                   │
        ├──────────────────────────────────┤
        │ <POINT (935996.821 191376.75)>   │
        │ <POINT (1031085.719 164018.754)> │
        │ <POINT (1026452.617 254265.479)> │
        │ <POINT (990633.981 202959.782)>  │
        │ <POINT (931871.37 140681.351)>   │
        │ <POINT (964319.735 157998.936)>  │
        │ <POINT (1006496.679 216719.218)> │
        │ <POINT (1005551.571 222936.088)> │
        │ <POINT (1043002.677 212969.849)> │
        │ <POINT (1042223.605 186706.496)> │
        │ …                                │
        └──────────────────────────────────┘
        """
        return ops.GeoPoint(self, right).to_expr()


@public
class NumericScalar(Scalar, NumericValue):
    pass


@public
class NumericColumn(Column, NumericValue):
    def std(
        self,
        where: ir.BooleanValue | None = None,
        how: Literal["sample", "pop"] = "sample",
    ) -> NumericScalar:
        """Return the standard deviation of a numeric column.

        Parameters
        ----------
        where
            Filter
        how
            Sample or population standard deviation

        Returns
        -------
        NumericScalar
            Standard deviation of `arg`
        """
        return ops.StandardDev(
            self, how=how, where=self._bind_to_parent_table(where)
        ).to_expr()

    def var(
        self,
        where: ir.BooleanValue | None = None,
        how: Literal["sample", "pop"] = "sample",
    ) -> NumericScalar:
        """Return the variance of a numeric column.

        Parameters
        ----------
        where
            Filter
        how
            Sample or population variance

        Returns
        -------
        NumericScalar
            Standard deviation of `arg`
        """
        return ops.Variance(
            self, how=how, where=self._bind_to_parent_table(where)
        ).to_expr()

    def corr(
        self,
        right: NumericColumn,
        where: ir.BooleanValue | None = None,
        how: Literal["sample", "pop"] = "sample",
    ) -> NumericScalar:
        """Return the correlation of two numeric columns.

        Parameters
        ----------
        right
            Numeric column
        where
            Filter
        how
            Population or sample correlation

        Returns
        -------
        NumericScalar
            The correlation of `left` and `right`
        """
        return ops.Correlation(
            self,
            self._bind_to_parent_table(right),
            how=how,
            where=self._bind_to_parent_table(where),
        ).to_expr()

    def cov(
        self,
        right: NumericColumn,
        where: ir.BooleanValue | None = None,
        how: Literal["sample", "pop"] = "sample",
    ) -> NumericScalar:
        """Return the covariance of two numeric columns.

        Parameters
        ----------
        right
            Numeric column
        where
            Filter
        how
            Population or sample covariance

        Returns
        -------
        NumericScalar
            The covariance of `self` and `right`
        """
        return ops.Covariance(
            self,
            self._bind_to_parent_table(right),
            how=how,
            where=self._bind_to_parent_table(where),
        ).to_expr()

    def mean(
        self,
        where: ir.BooleanValue | None = None,
    ) -> NumericScalar:
        """Return the mean of a numeric column.

        Parameters
        ----------
        where
            Filter

        Returns
        -------
        NumericScalar
            The mean of the input expression
        """
        # TODO(kszucs): remove the alias from the reduction method in favor
        # of default name generated by ops.Value operations
        return ops.Mean(self, where=self._bind_to_parent_table(where)).to_expr()

    def cummean(self, *, where=None, group_by=None, order_by=None) -> NumericColumn:
        """Return the cumulative mean of the input."""
        return self.mean(where=where).over(
            ibis.cumulative_window(group_by=group_by, order_by=order_by)
        )

    def sum(
        self,
        where: ir.BooleanValue | None = None,
    ) -> NumericScalar:
        """Return the sum of a numeric column.

        Parameters
        ----------
        where
            Filter

        Returns
        -------
        NumericScalar
            The sum of the input expression
        """
        return ops.Sum(self, where=self._bind_to_parent_table(where)).to_expr()

    def cumsum(self, *, where=None, group_by=None, order_by=None) -> NumericColumn:
        """Return the cumulative sum of the input."""
        return self.sum(where=where).over(
            ibis.cumulative_window(group_by=group_by, order_by=order_by)
        )

    def bucket(
        self,
        buckets: Sequence[int],
        closed: Literal["left", "right"] = "left",
        close_extreme: bool = True,
        include_under: bool = False,
        include_over: bool = False,
    ) -> ir.IntegerColumn:
        """Compute a discrete binning of a numeric array.

        Parameters
        ----------
        buckets
            List of buckets
        closed
            Which side of each interval is closed. For example:

            ```python
            buckets = [0, 100, 200]
            closed = "left"  # 100 falls in 2nd bucket
            closed = "right"  # 100 falls in 1st bucket
            ```
        close_extreme
            Whether the extreme values fall in the last bucket
        include_over
            Include values greater than the last bucket in the last bucket
        include_under
            Include values less than the first bucket in the first bucket

        Returns
        -------
        IntegerColumn
            A categorical column expression
        """
        return ops.Bucket(
            self,
            buckets,
            closed=closed,
            close_extreme=close_extreme,
            include_under=include_under,
            include_over=include_over,
        ).to_expr()

    def histogram(
        self,
        nbins: int | None = None,
        binwidth: float | None = None,
        base: float | None = None,
        eps: float = 1e-13,
    ):
        """Compute a histogram with fixed width bins.

        Parameters
        ----------
        nbins
            If supplied, will be used to compute the binwidth
        binwidth
            If not supplied, computed from the data (actual max and min values)
        base
            The value of the first histogram bin. Defaults to the minimum value
            of `column`.
        eps
            Allowed floating point epsilon for histogram base

        Returns
        -------
        Column
            Bucketed column
        """

        if nbins is not None and binwidth is not None:
            raise ValueError(
                f"Cannot pass both `nbins` (got {nbins}) and `binwidth` (got {binwidth})"
            )

        if base is None:
            base = self.min() - eps

        if binwidth is None:
            if nbins is None:
                raise ValueError("`nbins` is required if `binwidth` is not provided")

            binwidth = (self.max() - base) / nbins

        if nbins is None:
            nbins = ((self.max() - base) / binwidth).ceil()

        return ((self - base) / binwidth).floor().clip(-1, nbins - 1)

    def approx_quantile(
        self,
        quantile: float | ir.NumericValue | Sequence[ir.NumericValue | float],
        where: ir.BooleanValue | None = None,
    ) -> NumericScalar:
        """Compute one or more approximate quantiles of a column.

        ::: {.callout-note}
        ## The result may or may not be exact

        Whether the result is an approximation depends on the backend.
        :::

        Parameters
        ----------
        quantile
            `0 <= quantile <= 1`, or an array of such values
            indicating the quantile or quantiles to compute
        where
            Boolean filter for input values

        Returns
        -------
        Scalar
            Quantile of the input

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.examples.penguins.fetch()

        Compute the approximate 0.50 quantile of `bill_depth_mm`.

        >>> t.bill_depth_mm.approx_quantile(0.50)
        ┌────────┐
        │ 17.318 │
        └────────┘

        Compute multiple approximate quantiles in one call - in this case the
        result is an array.

        >>> t.bill_depth_mm.approx_quantile([0.25, 0.75])
        ┌────────────────────────┐
        │ [15.565625, 18.671875] │
        └────────────────────────┘
        """
        if isinstance(quantile, Sequence):
            op = ops.ApproxMultiQuantile
        else:
            op = ops.ApproxQuantile
        return op(self, quantile, where=self._bind_to_parent_table(where)).to_expr()


@public
class IntegerValue(NumericValue):
    def as_timestamp(
        self,
        unit: Literal["s", "ms", "us"],
    ) -> ir.TimestampValue:
        """Convert an integral UNIX timestamp to a timestamp expression.

        Parameters
        ----------
        unit
            The resolution of `arg`

        Returns
        -------
        TimestampValue
            `self` converted to a timestamp
        """
        return ops.TimestampFromUNIX(self, unit).to_expr()

    def as_interval(
        self,
        unit: Literal["Y", "M", "W", "D", "h", "m", "s", "ms", "us", "ns"] = "s",
    ) -> ir.IntervalValue:
        """Convert an integer to an interval.

        Parameters
        ----------
        unit
            Unit for the resulting interval

        Returns
        -------
        IntervalValue
            An interval in units of `unit`
        """
        return ops.IntervalFromInteger(self, unit).to_expr()

    @deprecated(as_of="10.0", instead="use as_timestamp() instead")
    def to_timestamp(
        self,
        unit: Literal["s", "ms", "us"] = "s",
    ) -> ir.TimestampValue:
        return self.as_timestamp(unit=unit)

    @deprecated(as_of="10.0", instead="use as_interval() instead")
    def to_interval(
        self,
        unit: Literal["Y", "M", "W", "D", "h", "m", "s", "ms", "us", "ns"] = "s",
    ) -> ir.IntervalValue:
        return self.as_interval(unit=unit)

    def convert_base(
        self,
        from_base: IntegerValue,
        to_base: IntegerValue,
    ) -> IntegerValue:
        """Convert an integer from one base to another.

        Parameters
        ----------
        from_base
            Numeric base of expression
        to_base
            New base

        Returns
        -------
        IntegerValue
            Converted expression
        """
        return ops.BaseConvert(self, from_base, to_base).to_expr()

    def __and__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise and `self` with `other`."""
        return _binop(ops.BitwiseAnd, self, other)

    __rand__ = __and__

    def __or__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise or `self` with `other`."""
        return _binop(ops.BitwiseOr, self, other)

    __ror__ = __or__

    def __xor__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise xor `self` with `other`."""
        return _binop(ops.BitwiseXor, self, other)

    __rxor__ = __xor__

    def __lshift__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise left shift `self` with `other`."""
        return _binop(ops.BitwiseLeftShift, self, other)

    def __rlshift__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise left shift `self` with `other`."""
        return _binop(ops.BitwiseLeftShift, other, self)

    def __rshift__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise right shift `self` with `other`."""
        return _binop(ops.BitwiseRightShift, self, other)

    def __rrshift__(self, other: IntegerValue) -> IntegerValue:
        """Bitwise right shift `self` with `other`."""
        return _binop(ops.BitwiseRightShift, other, self)

    def __invert__(self) -> IntegerValue:
        """Bitwise not of `self`.

        Returns
        -------
        IntegerValue
            Inverted bits of `self`.
        """
        try:
            node = ops.BitwiseNot(self)
        except (XorqTypeError, NotImplementedError):
            return NotImplemented
        else:
            return node.to_expr()

    def label(self, labels: Iterable[str], nulls: str | None = None) -> ir.StringValue:
        """Label a set of integer values with strings.

        Parameters
        ----------
        labels
            An iterable of string labels. Each integer value in `self` will be mapped to
            a value in `labels`.
        nulls
            String label to use for `NULL` values

        Returns
        -------
        StringValue
            `self` labeled with `labels`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> t = ibis.memtable({"a": [0, 1, 0, 2]})
        >>> t.select(t.a, labeled=t.a.label(["a", "b", "c"]))
        ┏━━━━━━━┳━━━━━━━━━┓
        ┃ a     ┃ labeled ┃
        ┡━━━━━━━╇━━━━━━━━━┩
        │ int64 │ string  │
        ├───────┼─────────┤
        │     0 │ a       │
        │     1 │ b       │
        │     0 │ a       │
        │     2 │ c       │
        └───────┴─────────┘
        """
        return (
            functools.reduce(
                lambda stmt, inputs: stmt.when(*inputs), enumerate(labels), self.case()
            )
            .else_(nulls)
            .end()
        )


@public
class IntegerScalar(NumericScalar, IntegerValue):
    pass


@public
class IntegerColumn(NumericColumn, IntegerValue):
    def bit_and(self, where: ir.BooleanValue | None = None) -> IntegerScalar:
        """Aggregate the column using the bitwise and operator."""
        return ops.BitAnd(self, where=self._bind_to_parent_table(where)).to_expr()

    def bit_or(self, where: ir.BooleanValue | None = None) -> IntegerScalar:
        """Aggregate the column using the bitwise or operator."""
        return ops.BitOr(self, where=self._bind_to_parent_table(where)).to_expr()

    def bit_xor(self, where: ir.BooleanValue | None = None) -> IntegerScalar:
        """Aggregate the column using the bitwise exclusive or operator."""
        return ops.BitXor(self, where=self._bind_to_parent_table(where)).to_expr()


@public
class FloatingValue(NumericValue):
    def isnan(self) -> ir.BooleanValue:
        """Return whether the value is NaN."""
        return ops.IsNan(self).to_expr()

    def isinf(self) -> ir.BooleanValue:
        """Return whether the value is infinity."""
        return ops.IsInf(self).to_expr()


@public
class FloatingScalar(NumericScalar, FloatingValue):
    pass


@public
class FloatingColumn(NumericColumn, FloatingValue):
    pass


@public
class DecimalValue(NumericValue):
    pass


@public
class DecimalScalar(NumericScalar, DecimalValue):
    pass


@public
class DecimalColumn(NumericColumn, DecimalValue):
    pass



import pytest

import xorq.vendor.ibis as ibis


@pytest.fixture
def orders():
    return ibis.table(
        {
            "o_orderkey": "int32",
            "o_orderpriority": "string",
            "o_custkey": "int32",
            "o_orderdate": "date",
        },
        name="orders",
    )


@pytest.fixture
def supplier():
    return ibis.table(
        {
            "s_suppkey": "int32",
            "s_nationkey": "int32",
        },
        name="supplier",
    )


@pytest.fixture
def lineitem():
    return ibis.table(
        {
            "l_orderkey": "int64",
            "l_suppkey": "int32",
            "l_shipdate": "date",
            "l_extendedprice": "decimal(15,2)",
            "l_discount": "decimal(15,2)",
        },
        name="lineitem",
    )


@pytest.fixture
def customer():
    return ibis.table(
        {
            "c_custkey": "int32",
            "c_nationkey": "int32",
        },
        name="customer",
    )


@pytest.fixture
def nation():
    return ibis.table(
        {
            "n_nationkey": "int32",
            "n_name": "string",
        },
        name="nation",
    )


# Minimal test mimicking h07 join chain with self-reference in projection
def test_minimal_joinchain_self_reference(
    compiler, orders, supplier, lineitem, customer, nation
):
    q = supplier.join(lineitem, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(customer, customer.c_custkey == orders.o_custkey)
    n1 = nation
    n2 = nation.view()
    q = q.join(n1, supplier.s_nationkey == n1.n_nationkey)
    q = q.join(n2, customer.c_nationkey == n2.n_nationkey)
    q = q.projection(
        {
            "supp_nation": n1.n_name,
            "cust_nation": n2.n_name,
            "l_shipdate": lineitem.l_shipdate,
            "l_extendedprice": lineitem.l_extendedprice,
            "l_discount": lineitem.l_discount,
        }
    )
    q = q.filter(
        (
            ((q.cust_nation == "FRANCE") & (q.supp_nation == "GERMANY"))
            | ((q.cust_nation == "GERMANY") & (q.supp_nation == "FRANCE"))
        )
    )

    yaml_dict = compiler.to_yaml(q)
    q_roundtrip = compiler.from_yaml(yaml_dict)

    try:
        _ = q_roundtrip["cust_nation"]
    except Exception as e:
        pytest.fail(f"Accessing 'cust_nation' on the roundtrip query failed: {e}")



"""Test datetime operation translations."""

from datetime import datetime

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations.temporal as tm

from xorq.common.utils.ibis_utils import from_ibis


def test_date_extract():
    dt_expr = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))
    year = dt_expr.year()

    assert from_ibis(year) is not None
    assert from_ibis(dt_expr.month()) is not None
    assert from_ibis(dt_expr.day()) is not None


def test_time_extract():
    dt_expr = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))

    assert from_ibis(dt_expr.hour()) is not None
    assert from_ibis(dt_expr.minute()) is not None
    assert from_ibis(dt_expr.second()) is not None


def test_timestamp_arithmetic():
    ts = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))
    delta = ibis.interval(days=1)

    plus_day = ts + delta
    assert from_ibis(plus_day) is not None

    minus_day = ts - delta
    assert from_ibis(minus_day) is not None


def test_timestamp_diff():
    ts1 = ibis.literal(datetime(2024, 3, 14))
    ts2 = ibis.literal(datetime(2024, 3, 15))
    diff = ts2 - ts1

    assert from_ibis(diff) is not None


def test_temporal_unit_yaml():
    interval_date = ibis.literal(5, type=dt.Interval(unit=tm.DateUnit("D")))
    assert from_ibis(interval_date) is not None

    interval_time = ibis.literal(10, type=dt.Interval(unit=tm.TimeUnit("h")))
    assert from_ibis(interval_time) is not None



import pyarrow.parquet as pq
import pytest


ibis = pytest.importorskip("ibis")


@pytest.fixture(scope="session")
def ibis_con(parquet_dir, tmp_path_factory):
    path = tmp_path_factory.mktemp("data") / "small_alltypes.parquet"
    table = pq.read_table(parquet_dir / "functional_alltypes.parquet").slice(length=100)
    pq.write_table(table, path)

    conn = ibis.duckdb.connect()
    conn.read_parquet(path, table_name="alltypes")
    conn.read_parquet(parquet_dir / "astronauts.parquet", table_name="astronauts")
    return conn


@pytest.fixture(scope="session")
def ibis_alltypes(ibis_con):
    return ibis_con.table("alltypes")


@pytest.fixture(scope="session")
def ibis_astronauts(ibis_con):
    return ibis_con.table("astronauts")


@pytest.fixture
def t():
    return ibis.table(
        {
            "a": "int64",
            "b": "string",
            "c": "float64",
            "d": "timestamp",
            "e": "date",
        },
        name="test_table",
    )



from __future__ import annotations

import re
from abc import ABC, abstractmethod

import sqlglot as sg

import xorq.vendor.ibis.expr.datatypes as dt


fully_qualified_re = re.compile(r"(.*)\.(?:`(.*)`|(.*))")


# TODO(kszucs): the following classes should produce SQLGlot expressions


def is_fully_qualified(x):
    return bool(fully_qualified_re.search(x))


def _is_quoted(x):
    regex = re.compile(r"(?:`(.*)`|(.*))")
    quoted, _ = regex.match(x).groups()
    return quoted is not None


class Base(ABC):
    @property
    @abstractmethod
    def dialect(self): ...

    @abstractmethod
    def compile(self): ...

    def quote(self, ident):
        return sg.to_identifier(ident, quoted=True).sql(dialect=self.dialect)

    def scoped_name(
        self, obj_name: str, database: str | None = None, catalog: str | None = None
    ) -> str:
        if is_fully_qualified(obj_name):
            return obj_name
        if _is_quoted(obj_name):
            obj_name = obj_name[1:-1]
        return sg.table(obj_name, db=database, catalog=catalog, quoted=True).sql(
            dialect=self.dialect
        )

    @abstractmethod
    def format_dtype(self, dtype): ...

    def format_schema(self, schema):
        elements = [
            f"{self.quote(name)} {self.format_dtype(t)}"
            for name, t in zip(schema.names, schema.types)
        ]
        return "({})".format(",\n ".join(elements))

    def format_partition(self, partition, partition_schema):
        def _format_partition_kv(k, v, type):
            if type == dt.string:
                value_formatted = f'"{v}"'
            else:
                value_formatted = str(v)

            return f"{k}={value_formatted}"

        tokens = []
        if isinstance(partition, dict):
            for name in partition_schema:
                if name in partition:
                    tok = _format_partition_kv(
                        name, partition[name], partition_schema[name]
                    )
                else:
                    # dynamic partitioning
                    tok = name
                tokens.append(tok)
        else:
            for name, value in zip(partition_schema, partition):
                tok = _format_partition_kv(name, value, partition_schema[name])
                tokens.append(tok)

        return "PARTITION ({})".format(", ".join(tokens))


class DML(Base):
    pass


class DDL(Base):
    pass


class CreateDDL(DDL):
    def _if_exists(self):
        return "IF NOT EXISTS " if self.can_exist else ""


class DropObject(DDL):
    def __init__(self, must_exist=True):
        self.must_exist = must_exist

    def compile(self):
        if_exists = "" if self.must_exist else "IF EXISTS "
        object_name = self._object_name()
        return f"DROP {self._object_type} {if_exists}{object_name}"


class DropFunction(DropObject):
    def __init__(self, name, inputs, must_exist=True, aggregate=False, database=None):
        super().__init__(must_exist=must_exist)
        self.name = name
        self.inputs = tuple(map(dt.dtype, inputs))
        self.must_exist = must_exist
        self.aggregate = aggregate
        self.database = database

    def _object_name(self):
        return self.name

    def compile(self):
        tokens = ["DROP"]
        if self.aggregate:
            tokens.append("AGGREGATE")
        tokens.append("FUNCTION")
        if not self.must_exist:
            tokens.append("IF EXISTS")

        tokens.append(self._impala_signature())
        return " ".join(tokens)



import pytest

from xorq.common.utils.ibis_utils import from_ibis


BASE_URL = "https://pub-a45a6a332b4646f2a6f44775695c64df.r2.dev"


@pytest.mark.slow(level=3)
def test_complex_expr_from_boring_semantic_layer(ibis_con):
    flights = ibis_con.read_parquet(f"{BASE_URL}/flights.parquet")
    carriers = ibis_con.read_parquet(f"{BASE_URL}/carriers.parquet")

    # Join flights with carrier information
    flights_with_carriers = flights.inner_join(
        carriers,
        (flights.carrier == carriers.code),
    ).select(
        flights.carrier,
        flights.origin,
        flights.destination,
        flights.flight_num,
        flights.flight_time,
        flights.tail_num,
        flights.dep_time,
        flights.arr_time,
        flights.dep_delay,
        flights.arr_delay,
        flights.taxi_out,
        flights.taxi_in,
        flights.distance,
        flights.cancelled,
        flights.diverted,
        flights.id2,
        carriers.code,
        carriers.name,
        carriers.nickname,
    )

    # Calculate statistics by carrier
    distance_sum_expr = flights_with_carriers.distance.sum()
    carrier_stats = flights_with_carriers.aggregate(
        [
            flights_with_carriers.count().name("flight_count"),
            distance_sum_expr.name("total_distance"),
        ],
        by=[flights_with_carriers.nickname],
    )

    # Calculate overall total distance
    overall_totals = flights_with_carriers.aggregate(
        [distance_sum_expr.name("total_distance")]
    )

    # Join carrier stats with overall totals
    carrier_stats_with_totals = carrier_stats.cross_join(overall_totals).select(
        carrier_stats.nickname,
        carrier_stats.flight_count,
        carrier_stats.total_distance,
        overall_totals.total_distance.name("overall_total_distance"),
    )

    # Calculate distance share percentage for each carrier
    carrier_distance_shares = carrier_stats_with_totals.select(
        carrier_stats_with_totals.nickname,
        carrier_stats_with_totals.flight_count,
        (
            (
                carrier_stats_with_totals.total_distance.cast("float64")
                / carrier_stats_with_totals.overall_total_distance.cast("float64")
            )
            * 100
        ).name("distance_share_percentage"),
    )

    # Get top 10 carriers by distance share
    top_carriers_by_distance_share = carrier_distance_shares.order_by(
        carrier_distance_shares.distance_share_percentage.desc()
    ).limit(10)

    xorq_expr = from_ibis(top_carriers_by_distance_share)

    assert xorq_expr is not None
    assert not xorq_expr.execute().empty



from operator import methodcaller

import dask
import pyarrow.parquet as pq
import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)
from xorq.tests.util import assert_frame_equal


def test_can_connect(sqlite_con):
    assert sqlite_con is not None
    assert sqlite_con.list_tables() is not None


def test_read_record_batch_reader(sqlite_con, astronauts_parquet_path):
    reader = pq.read_table(astronauts_parquet_path).to_reader()
    expr = sqlite_con.read_record_batches(reader).filter(xo._.number == 104)
    assert not expr.execute().empty


def test_can_into_backend(sqlite_con, astronauts_parquet_path):
    astronauts_table = pq.read_table(astronauts_parquet_path)
    reader = astronauts_table.to_reader()
    astronauts = sqlite_con.read_record_batches(reader, table_name="astronauts")
    ddb = xo.duckdb.connect()

    expr = astronauts.join(
        ddb.read_in_memory(
            astronauts_table.to_pandas(), table_name="astronauts"
        ).into_backend(sqlite_con, name="ddb_astronauts"),
        "id",
    ).filter(xo._.number == 104)

    assert not expr.execute().empty


def test_read_parquet(sqlite_con, astronauts_parquet_path):
    t = sqlite_con.read_parquet(astronauts_parquet_path)
    assert not t.execute().empty


def test_can_cache(sqlite_con, astronauts_parquet_path):
    ddb = xo.duckdb.connect()
    astronauts = ddb.read_parquet(astronauts_parquet_path)

    expr = (
        astronauts.cache(SourceCache.from_kwargs(source=sqlite_con))
        .filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
    )

    actual = expr.execute()
    expected = (
        astronauts.filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
        .execute()
    )

    assert_frame_equal(expected, actual)


def test_can_be_cached(sqlite_con, astronauts_parquet_path):
    astronauts = sqlite_con.read_parquet(astronauts_parquet_path)
    expr = (
        astronauts.cache(SourceCache.from_kwargs(source=xo.duckdb.connect()))
        .filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
    )

    actual = expr.execute()
    expected = (
        astronauts.filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
        .execute()
    )

    assert_frame_equal(expected, actual)


@pytest.mark.parametrize("collect", ["to_pyarrow", "to_pyarrow_batches", "execute"])
def test_can_collect(sqlite_con, astronauts_parquet_path, collect):
    astronauts = sqlite_con.read_parquet(astronauts_parquet_path)
    expr = (
        astronauts.filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
        .mutate(add_1=xo._.number + 1, clean_name=xo._.name.strip())
    )
    assert methodcaller(collect)(expr) is not None


def test_can_outo_backend_and_tokenize(sqlite_con, astronauts_parquet_path):
    ddb = xo.duckdb.connect()

    astronauts = sqlite_con.read_parquet(astronauts_parquet_path)
    expr = (
        astronauts.filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
        .mutate(add_1=xo._.number + 1, clean_name=xo._.name.strip())
        .into_backend(ddb, name="ddb_astronauts")
    )

    assert dask.base.tokenize(expr) is not None
    assert not expr.execute().empty


@pytest.mark.parametrize("file_format", ["csv", "parquet"])
def test_can_deferred_read(sqlite_con, file_format, request):
    read = methodcaller(
        f"deferred_read_{file_format}",
        request.getfixturevalue(f"astronauts_{file_format}_path"),
        con=sqlite_con,
    )(xo)
    assert not read.execute().empty


def test_sqlite_snapshot(con_snapshot):
    con_snapshot(xo.sqlite.connect())


def test_cross_source_snapshot(con_cross_source_snapshot):
    con_cross_source_snapshot(xo.sqlite.connect())


@pytest.mark.parametrize(
    "cls",
    [ParquetSnapshotCache, ParquetCache, SourceSnapshotCache, SourceCache],
)
def test_cache_find_backend(cls, con_cache_find_backend):
    con_cache_find_backend(cls, xo.sqlite.connect())


def test_can_train_test_split(sqlite_con, astronauts_parquet_path):
    t = sqlite_con.read_parquet(astronauts_parquet_path)
    (train, test) = xo.train_test_splits(t, 0.2)
    assert train.execute() is not None
    assert test.execute() is not None



"""Test datetime operation translations."""

from datetime import datetime

import pytest

import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations.temporal as tm
from xorq.ibis_yaml.tests.conftest import get_dtype_yaml


def test_date_extract(compiler):
    dt_expr = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))

    year = dt_expr.year()
    year_yaml = compiler.to_yaml(year)
    expression = year_yaml["expression"]
    dtype_yaml = get_dtype_yaml(year_yaml, expression)
    assert expression["op"] == "ExtractYear"
    assert expression["arg"]["value"] == "2024-03-14T15:09:26"
    assert dtype_yaml["type"] == "Int32"
    roundtrip_year = compiler.from_yaml(year_yaml)
    assert roundtrip_year.equals(year)

    month = dt_expr.month()
    month_yaml = compiler.to_yaml(month)
    expression = month_yaml["expression"]
    dtype_yaml = get_dtype_yaml(month_yaml, expression)
    assert expression["op"] == "ExtractMonth"
    roundtrip_month = compiler.from_yaml(month_yaml)
    assert roundtrip_month.equals(month)

    day = dt_expr.day()
    day_yaml = compiler.to_yaml(day)
    expression = day_yaml["expression"]
    assert expression["op"] == "ExtractDay"
    roundtrip_day = compiler.from_yaml(day_yaml)
    assert roundtrip_day.equals(day)


def test_time_extract(compiler):
    dt_expr = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))

    hour = dt_expr.hour()
    hour_yaml = compiler.to_yaml(hour)
    hour_expression = hour_yaml["expression"]
    dtype_yaml = get_dtype_yaml(hour_yaml, hour_expression)
    assert hour_expression["op"] == "ExtractHour"
    assert hour_expression["arg"]["value"] == "2024-03-14T15:09:26"
    assert dtype_yaml["type"] == "Int32"
    roundtrip_hour = compiler.from_yaml(hour_yaml)
    assert roundtrip_hour.equals(hour)

    minute = dt_expr.minute()
    minute_yaml = compiler.to_yaml(minute)
    minute_expression = minute_yaml["expression"]
    assert minute_expression["op"] == "ExtractMinute"
    roundtrip_minute = compiler.from_yaml(minute_yaml)
    assert roundtrip_minute.equals(minute)

    second = dt_expr.second()
    second_yaml = compiler.to_yaml(second)
    second_expression = second_yaml["expression"]
    assert second_expression["op"] == "ExtractSecond"
    roundtrip_second = compiler.from_yaml(second_yaml)
    assert roundtrip_second.equals(second)


def test_timestamp_arithmetic(compiler):
    ts = ibis.literal(datetime(2024, 3, 14, 15, 9, 26))
    delta = ibis.interval(days=1)

    plus_day = ts + delta
    yaml_dict = compiler.to_yaml(plus_day)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert expression["op"] == "TimestampAdd"
    assert dtype_yaml["type"] == "Timestamp"
    assert get_dtype_yaml(yaml_dict, expression["right"])["type"] == "Interval"
    roundtrip_plus = compiler.from_yaml(yaml_dict)
    assert roundtrip_plus.equals(plus_day)

    minus_day = ts - delta
    yaml_dict = compiler.to_yaml(minus_day)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert expression["op"] == "TimestampSub"
    assert dtype_yaml["type"] == "Timestamp"
    assert get_dtype_yaml(yaml_dict, expression["right"])["type"] == "Interval"
    roundtrip_minus = compiler.from_yaml(yaml_dict)
    assert roundtrip_minus.equals(minus_day)


def test_timestamp_diff(compiler):
    ts1 = ibis.literal(datetime(2024, 3, 14))
    ts2 = ibis.literal(datetime(2024, 3, 15))
    diff = ts2 - ts1
    yaml_dict = compiler.to_yaml(diff)
    expression = yaml_dict["expression"]
    dtype_yaml = get_dtype_yaml(yaml_dict, expression)
    assert expression["op"] == "TimestampDiff"
    assert dtype_yaml["type"] == "Interval"
    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert roundtrip_expr.equals(diff)


def test_temporal_unit_yaml(compiler):
    interval_date = ibis.literal(5, type=dt.Interval(unit=tm.DateUnit("D")))
    yaml_date = compiler.to_yaml(interval_date)
    expression_date = yaml_date["expression"]
    dtype_yaml = get_dtype_yaml(yaml_date, expression_date)
    assert dtype_yaml["type"] == "Interval"
    assert dtype_yaml["unit"]["name"] == "DateUnit"
    assert dtype_yaml["unit"]["value"] == "D"
    roundtrip_date = compiler.from_yaml(yaml_date)
    assert roundtrip_date.equals(interval_date)

    interval_time = ibis.literal(10, type=dt.Interval(unit=tm.TimeUnit("h")))
    yaml_time = compiler.to_yaml(interval_time)
    expression_time = yaml_time["expression"]
    dtype_yaml = get_dtype_yaml(yaml_time, expression_time)
    assert dtype_yaml["type"] == "Interval"
    assert dtype_yaml["unit"]["name"] == "TimeUnit"
    assert dtype_yaml["unit"]["value"] == "h"
    roundtrip_time = compiler.from_yaml(yaml_time)
    assert roundtrip_time.equals(interval_time)


def test_date_truncate(compiler, t):
    """Test that DateTruncate with DateUnit serializes and roundtrips correctly."""
    truncated = t.e.truncate("M")
    yaml_dict = compiler.to_yaml(truncated)
    expression = yaml_dict["expression"]
    assert expression["op"] == "DateTruncate"
    assert expression["unit"]["op"] == "IntervalUnit"
    assert expression["unit"]["name"] == "DateUnit"
    assert expression["unit"]["value"] == "M"
    roundtrip = compiler.from_yaml(yaml_dict)
    assert roundtrip.equals(truncated)


def test_time_truncate(compiler):
    """Test that TimeTruncate with TimeUnit serializes and roundtrips correctly."""
    t = ibis.table({"t": "time"}, name="time_table")
    truncated = t.t.truncate("h")
    yaml_dict = compiler.to_yaml(truncated)
    expression = yaml_dict["expression"]
    assert expression["op"] == "TimeTruncate"
    assert expression["unit"]["op"] == "IntervalUnit"
    assert expression["unit"]["name"] == "TimeUnit"
    assert expression["unit"]["value"] == "h"
    roundtrip = compiler.from_yaml(yaml_dict)
    assert roundtrip.equals(truncated)


@pytest.mark.parametrize(
    "unit_cls,unit_value,expected_name",
    [
        (tm.DateUnit, "Y", "DateUnit"),
        (tm.DateUnit, "M", "DateUnit"),
        (tm.DateUnit, "D", "DateUnit"),
        (tm.TimeUnit, "h", "TimeUnit"),
        (tm.TimeUnit, "m", "TimeUnit"),
        (tm.TimeUnit, "s", "TimeUnit"),
        (tm.TimestampUnit, "s", "TimestampUnit"),
        (tm.TimestampUnit, "ms", "TimestampUnit"),
        (tm.IntervalUnit, "Y", "DateUnit"),
        (tm.IntervalUnit, "h", "TimeUnit"),
    ],
)
def test_temporal_unit_direct_serialization(
    compiler, unit_cls, unit_value, expected_name
):
    """Test that DateUnit, TimeUnit, and TimestampUnit enums serialize directly."""
    from xorq.ibis_yaml.common import TranslationContext, translate_to_yaml

    ctx = TranslationContext()
    unit = unit_cls(unit_value)
    result = translate_to_yaml(unit, ctx)
    assert result["op"] == "IntervalUnit"
    assert result["name"] == expected_name
    assert result["value"] == unit_value



from __future__ import annotations

from xorq.vendor.ibis.backends.sql.compilers.postgres import (
    PostgresCompiler as IbisPostgresCompiler,
)


_UNIX_EPOCH = "1970-01-01T00:00:00Z"


class PostgresCompiler(IbisPostgresCompiler):
    __slots__ = ()


compiler = PostgresCompiler()



import pytest

import xorq.api as xo


def test_register_read_csv(csv_dir):
    # this will use ls.options.backend: do we want to clear it out between function invocations?
    api_batting = xo.read_csv(csv_dir / "batting.csv", table_name="api_batting")
    result = xo.execute(api_batting)

    assert result is not None


def test_register_read_parquet(parquet_dir):
    # this will use ls.options.backend: do we want to clear it out between function invocations?
    api_batting = xo.read_parquet(
        parquet_dir / "batting.parquet", table_name="api_batting"
    )
    result = xo.execute(api_batting)

    assert result is not None


@pytest.mark.xfail(reason="No purpose with no registration api")
def test_executed_on_original_backend(parquet_dir, csv_dir, mocker):
    con = xo.config._backend_init()
    spy = mocker.spy(con, "execute")

    parquet_table = xo.read_parquet(parquet_dir / "batting.parquet")[
        lambda t: t.yearID == 2015
    ]

    csv_table = xo.read_csv(csv_dir / "batting.csv")[lambda t: t.yearID == 2014]

    expr = parquet_table.join(
        csv_table,
        "playerID",
    )

    assert xo.execute(expr) is not None
    assert spy.call_count == 1


@pytest.mark.parametrize(
    ("with_repartition_file_scans", "keep_partition_by_columns"),
    [(True, True), (True, False), (False, True), (False, False)],
)
def test_with_config(
    with_repartition_file_scans, keep_partition_by_columns, parquet_dir
):
    import pandas as pd

    from xorq.api import SessionConfig

    session_config = (
        SessionConfig()
        .with_repartition_file_scans(with_repartition_file_scans)
        .set(
            "datafusion.execution.keep_partition_by_columns",
            str(keep_partition_by_columns).lower(),
        )
    )

    con = xo.connect(session_config=session_config)

    expr = con.read_parquet(parquet_dir / "batting.parquet").limit(10)
    result = expr.execute()

    assert isinstance(result, pd.DataFrame)
    assert len(result) == 10


def test_cases(alltypes):
    expr = xo.cases((alltypes["bool_col"], alltypes["string_col"]), else_=None).substr(
        0, 2
    )

    assert expr.execute() is not None
